import logging
import time
from typing import Dict, List

import pandas as pd

from webviz_subsurface._utils.ensemble_summary_provider_set import (
    EnsembleSummaryProviderSet,
)


# -------------------
def get_df_smry(
    input_provider_set: EnsembleSummaryProviderSet,
    ensemble_names: List[str],
    ens_vectors: Dict[str, List[str]],
    ens_realizations: Dict[str, List[int]],
    selector_realizations: List[int],
    selector_well_names: List[str],
    selector_phases: List[str],
    selector_dates: List[str],
) -> pd.DataFrame:
    """Return dataframe filtered on ensemble names, realizations,
    well names, phases and dates selectors."""

    start_time = time.time()

    dfs = []

    for ens_name in ensemble_names:

        df_smry = _get_filtered_df(
            input_provider_set,
            ens_name,
            ens_vectors,
            ens_realizations,
            selector_realizations,
            selector_well_names,
            selector_phases,
            selector_dates,
        )

        dfs.append(df_smry)

    logging.debug(
        f"\n--- get_df_smry --- Total time: {time.time() - start_time} seconds.\n"
    )

    return pd.concat(dfs)


# --------------------------------
def get_df_diff(
    df_smry: pd.DataFrame,
    obs_error_weight: float = 0,
    weight_reduction_factor_oil: float = 1,
    weight_reduction_factor_wat: float = 1,
    weight_reduction_factor_gas: float = 1,
    misfit_exponent: float = 1.0,
    relative_diff: bool = False,
) -> pd.DataFrame:
    """Return dataframe with diff (sim-obs) for all data.
    Return empty dataframe if no realizations included."""

    start_time = time.time()

    if df_smry.empty:
        return pd.DataFrame()

    df_diff = _build_df_diff(
        df_smry,
        obs_error_weight,
        weight_reduction_factor_oil,
        weight_reduction_factor_wat,
        weight_reduction_factor_gas,
        misfit_exponent,
        relative_diff,
    )

    logging.debug(
        f"\n--- get_df_diff --- Total time: {time.time() - start_time} seconds.\n"
    )
    return df_diff


# --------------------------------
# pylint: disable = too-many-locals
def get_df_diff_stat(df_diff: pd.DataFrame) -> pd.DataFrame:
    """Return dataframe with statistics of production difference
    across all realizations per ensemble, well and date.
    Return empty dataframe if no realizations included in df."""

    if df_diff.empty:
        return pd.DataFrame()

    ensembles = []
    dates = []
    wells = []
    diff_vectors = []
    diff_mean_values = []
    diff_std_values = []
    diff_p10_values = []
    diff_p90_values = []

    for ens_name, dframe in df_diff.groupby("ENSEMBLE"):
        for _date, ensdf in dframe.groupby("DATE"):

            for col in ensdf.columns:
                if ":" in col:
                    vector = col.split(":")[0]
                    if vector in [
                        "DIFF_WOPT",
                        "DIFF_WWPT",
                        "DIFF_WGPT",
                        "DIFF_GOPT",
                        "DIFF_GWPT",
                        "DIFF_GGPT",
                    ]:
                        well = col.split(":")[1]
                        wells.append(well)
                        dates.append(_date)
                        ensembles.append(ens_name)
                        diff_vectors.append(vector)
                        diff_mean_values.append(ensdf[col].mean())
                        diff_std_values.append(ensdf[col].std())
                        diff_p10_values.append(ensdf[col].quantile(0.9))
                        diff_p90_values.append(ensdf[col].quantile(0.1))

    df_stat = pd.DataFrame(
        data={
            "ENSEMBLE": ensembles,
            "WELL": wells,
            "VECTOR": diff_vectors,
            "DATE": dates,
            "DIFF_MEAN": diff_mean_values,
            "DIFF_STD": diff_std_values,
            "DIFF_P10": diff_p10_values,
            "DIFF_P90": diff_p90_values,
        }
    )

    return df_stat


# -- help functions -------------

# --------------------------------
def get_df_hist_avg(df_long: pd.DataFrame) -> pd.DataFrame:
    """Read long format dataframe and replace hist vectors (WOPTH, etc) column
    values for each realization with it's average value grouped by WELL, DATE and
    ENSEMBLE columns. The three groupby columns must be part of df_long.
    This function will only have an effect in cases where the history rates
    have uncertainty (varies per realization)."""

    for col in df_long.columns:
        if "PTH" in col:
            df_long[col] = df_long.groupby(["WELL", "ENSEMBLE", "DATE"])[col].transform(
                "mean"
            )
    return df_long


# --------------------------------
def _get_vector_types(phases: List[str]) -> List[str]:
    """Return summary vector types associated with phases."""

    vector_types = []
    if "Oil" in phases:
        vector_types.append("WOPT")
    if "Water" in phases:
        vector_types.append("WWPT")
    if "Gas" in phases:
        vector_types.append("WGPT")

    return vector_types


# --------------------------------
def _get_filtered_df(
    input_provider_set: EnsembleSummaryProviderSet,
    ens_name: str,
    ens_vectors: Dict[str, List[str]],
    ens_realizations: Dict[str, List[int]],
    selector_realizations: List[int],
    selector_well_names: List[str],
    selector_phases: List[str],
    selector_dates: List[str],
) -> List[str]:
    """Return filtered dataframe."""

    filtered_vector_types = _get_vector_types(selector_phases)

    filtered_vectors = []
    filtered_realizations = []

    for vector in ens_vectors[ens_name]:
        if (
            vector.split(":")[0] in filtered_vector_types
            and vector.split(":")[1] in selector_well_names
        ):
            hvector = vector.split(":")[0] + "H:" + vector.split(":")[1]
            filtered_vectors.append(vector)
            filtered_vectors.append(hvector)
    # logging.debug(f"Filtered vectors:\n{filtered_vectors}")

    filtered_realizations = [
        real
        for real in ens_realizations[ens_name]
        if real in set(selector_realizations)
    ]
    logging.debug(f"Filtered realizations:\n{filtered_realizations}")

    df = pd.DataFrame()
    if filtered_vectors and filtered_realizations:
        df = input_provider_set.provider(ens_name).get_vectors_df(
            filtered_vectors, None, filtered_realizations
        )

        df = df.astype({"DATE": "string"})
        df = df.loc[df["DATE"].isin(selector_dates)]  # --- apply date filter

        df["ENSEMBLE"] = ens_name

    return df


# --------------------------------
def _build_df_diff(
    df_smry: pd.DataFrame,
    obs_error_weight: float = 0,
    weight_reduction_factor_oil: float = 1,
    weight_reduction_factor_wat: float = 1,
    weight_reduction_factor_gas: float = 1,
    misfit_exponent: float = 1.0,
    relative_diff: bool = False,
) -> pd.DataFrame:
    """Calculate diffs (sim-obs) and return dataframe"""

    df_diff = df_smry[["ENSEMBLE", "DATE", "REAL"]].copy()

    for simvector in df_smry.columns:
        if "PT:" in simvector:
            vectortype, wellname = simvector.split(":")[0], simvector.split(":")[1]
            obsvector = vectortype + "H:" + wellname
            diff_col_name = "DIFF_" + vectortype + ":" + wellname
            if relative_diff:
                df_diff[diff_col_name] = (
                    (df_smry[simvector] - df_smry[obsvector])
                    / df_smry[obsvector].clip(1000)
                ) * 100
            elif obs_error_weight > 0:
                # obs error, including a lower bound (diminish very low values)
                obs_error = (obs_error_weight * df_smry[obsvector]).clip(1000)
                df_diff[diff_col_name] = (
                    (df_smry[simvector] - df_smry[obsvector]) / obs_error
                ) ** misfit_exponent
            elif obs_error_weight < 0:
                if vectortype == "WOPT":
                    weight_reduction = weight_reduction_factor_oil
                if vectortype == "WWPT":
                    weight_reduction = weight_reduction_factor_wat
                if vectortype == "WGPT":
                    weight_reduction = weight_reduction_factor_gas
                df_diff[diff_col_name] = (
                    (df_smry[simvector] - df_smry[obsvector]) / (weight_reduction)
                ) ** misfit_exponent
            else:
                df_diff[diff_col_name] = (
                    (df_smry[simvector] - df_smry[obsvector])
                ) ** misfit_exponent

    return df_diff



from pathlib import Path
from typing import List

import pandas as pd
import pytest

import webviz_subsurface._datainput.from_timeseries_cumulatives as from_cum


def get_data_df(testdata_folder: Path) -> pd.DataFrame:

    data_df = pd.read_csv(
        testdata_folder / "reek_test_data" / "aggregated_data" / "unsmry--monthly.csv"
    )
    data_df.DATE = data_df.DATE.astype(str)
    return data_df


def test_calc_from_cumulatives(testdata_folder: Path) -> None:
    # Includes monthly data, 10 reals x 4 ensembles, 3 years and 1 month (2000-01-01 to 2003-02-01)
    data_df = get_data_df(testdata_folder)

    ## Test single column key, FOPT as average rate avg_fopr, monthly
    calc_df = from_cum.calc_from_cumulatives(
        data=data_df,
        column_keys="FOPT",
        time_index="monthly",
        time_index_input="monthly",
        as_rate=True,
    )

    # Test real 0, iter-2
    real_data = data_df[(data_df["REAL"] == 0) & (data_df["ENSEMBLE"] == "iter-2")]
    real_calc = calc_df[(calc_df["REAL"] == 0) & (calc_df["ENSEMBLE"] == "iter-2")]

    assert real_calc[real_calc.DATE == "2000-01-01"]["AVG_FOPR"].values == (
        (
            real_data[real_data.DATE == "2000-02-01"]["FOPT"].values
            - real_data[real_data.DATE == "2000-01-01"]["FOPT"].values
        )
        / 31
    )

    assert real_calc[real_calc.DATE == "2002-05-01"]["AVG_FOPR"].values == (
        (
            real_data[real_data.DATE == "2002-06-01"]["FOPT"].values
            - real_data[real_data.DATE == "2002-05-01"]["FOPT"].values
        )
        / 31
    )

    ## Test multiple column keys, WOPT:OP_1 as average rate avg_fopr, monthly
    calc_df = from_cum.calc_from_cumulatives(
        data=data_df,
        column_keys=["WOPT:OP_1", "GOPT:OP"],
        time_index="yearly",
        time_index_input="monthly",
        as_rate=True,
    )
    # Test real 4, iter-0
    real_data = data_df[(data_df["REAL"] == 4) & (data_df["ENSEMBLE"] == "iter-0")]
    real_calc = calc_df[(calc_df["REAL"] == 4) & (calc_df["ENSEMBLE"] == "iter-0")]

    assert real_calc[real_calc.DATE == "2000-01-01"]["AVG_WOPR:OP_1"].values == (
        (
            real_data[real_data.DATE == "2001-01-01"]["WOPT:OP_1"].values
            - real_data[real_data.DATE == "2000-01-01"]["WOPT:OP_1"].values
        )
        / 366
    )

    assert real_calc[real_calc.DATE == "2002-01-01"]["AVG_GOPR:OP"].values == (
        (
            real_data[real_data.DATE == "2003-01-01"]["GOPT:OP"].values
            - real_data[real_data.DATE == "2002-01-01"]["GOPT:OP"].values
        )
        / 365
    )

    assert real_calc[real_calc.DATE == "2002-01-01"]["AVG_WOPR:OP_1"].values == (
        (
            real_data[real_data.DATE == "2003-01-01"]["WOPT:OP_1"].values
            - real_data[real_data.DATE == "2002-01-01"]["WOPT:OP_1"].values
        )
        / 365
    )

    assert real_calc[real_calc.DATE == "2001-01-01"]["AVG_GOPR:OP"].values == (
        (
            real_data[real_data.DATE == "2002-01-01"]["GOPT:OP"].values
            - real_data[real_data.DATE == "2001-01-01"]["GOPT:OP"].values
        )
        / 365
    )

    ## Test multiple column keys, WOPR_OP as average rate avg_fopr, monthly
    calc_df = from_cum.calc_from_cumulatives(
        data=data_df,
        column_keys=["WGPT:OP_2", "GWPT:OP"],
        time_index="monthly",
        time_index_input="monthly",
        as_rate=False,
    )
    # Test real 9, iter-0
    real_data = data_df[(data_df["REAL"] == 9) & (data_df["ENSEMBLE"] == "iter-0")]
    real_calc = calc_df[(calc_df["REAL"] == 9) & (calc_df["ENSEMBLE"] == "iter-0")]

    assert real_calc[real_calc.DATE == "2000-01-01"]["INTVL_WGPT:OP_2"].values == (
        real_data[real_data.DATE == "2000-01-01"]["WGPT:OP_2"].values
        - real_data[real_data.DATE == "2000-02-01"]["WGPT:OP_2"].values
    )

    assert real_calc[real_calc.DATE == "2002-05-01"]["INTVL_GWPT:OP"].values == (
        real_data[real_data.DATE == "2002-06-01"]["GWPT:OP"].values
        - real_data[real_data.DATE == "2002-05-01"]["GWPT:OP"].values
    )

    assert real_calc[real_calc.DATE == "2000-12-01"]["INTVL_WGPT:OP_2"].values == (
        real_data[real_data.DATE == "2001-01-01"]["WGPT:OP_2"].values
        - real_data[real_data.DATE == "2000-12-01"]["WGPT:OP_2"].values
    )

    assert real_calc[real_calc.DATE == "2002-02-01"]["INTVL_GWPT:OP"].values == (
        real_data[real_data.DATE == "2002-03-01"]["GWPT:OP"].values
        - real_data[real_data.DATE == "2002-02-01"]["GWPT:OP"].values
    )

    # Resample the data to yearly datapoints:
    data_df = data_df[
        data_df["DATE"].isin(["2000-01-01", "2001-01-01", "2002-01-01", "2003-01-01"])
    ]
    calc_df = from_cum.calc_from_cumulatives(
        data=data_df,
        column_keys=["WGPT:OP_2", "GWPT:OP"],
        time_index="yearly",
        time_index_input="yearly",
        as_rate=False,
    )
    # Test real 9, iter-0
    real_data = data_df[(data_df["REAL"] == 9) & (data_df["ENSEMBLE"] == "iter-0")]
    real_calc = calc_df[(calc_df["REAL"] == 9) & (calc_df["ENSEMBLE"] == "iter-0")]

    assert real_calc[real_calc.DATE == "2000-01-01"]["INTVL_WGPT:OP_2"].values == (
        real_data[real_data.DATE == "2001-01-01"]["WGPT:OP_2"].values
        - real_data[real_data.DATE == "2000-01-01"]["WGPT:OP_2"].values
    )

    assert real_calc[real_calc.DATE == "2002-01-01"]["INTVL_GWPT:OP"].values == (
        real_data[real_data.DATE == "2003-01-01"]["GWPT:OP"].values
        - real_data[real_data.DATE == "2002-01-01"]["GWPT:OP"].values
    )


@pytest.mark.parametrize(
    "column_keys,time_index,time_index_input,as_rate",
    [
        (["WGPT:OP_2", "GWPT:OP"], "monthly", "yearly", False),
        (["WGPT:OP_2", "GWPT:OP"], "daily", "monthly", True),
        (["WGPT:OP_2", "GWPT:OP"], "daily", "yearly", True),
    ],
)
def test_calc_from_cumulatives_errors(
    column_keys: List[str],
    time_index: str,
    time_index_input: str,
    as_rate: bool,
    testdata_folder: Path,
) -> None:
    data_df = get_data_df(testdata_folder)
    with pytest.raises(ValueError):
        # Is this variable necessary?
        # pylint: disable=unused-variable
        calc_df = from_cum.calc_from_cumulatives(
            data=data_df,
            column_keys=column_keys,
            time_index=time_index,
            time_index_input=time_index_input,
            as_rate=as_rate,
        )



import io
import json
import logging
from pathlib import Path
from typing import Any, Callable, Dict, List, Tuple

import pandas as pd
from webviz_config.webviz_store import webvizstore

from webviz_subsurface._datainput.fmu_input import scratch_ensemble


class WellAttributesModel:
    """Facilitates loading of a json file with well attributes.

    The file needs to follow the format below. The categorical attributes \
    are completely flexible.
    {
        "version" : "0.1",
        "wells" : [
            {
                "alias" : {
                    "eclipse" : "OP_1"
                },
                "attributes" : {
                    "mlt_singlebranch" : "mlt",
                    "structure" : "East",
                    "welltype" : "producer"
                },
                "name" : "OP_1"
            },
            {
                "alias" : {
                    "eclipse" : "GI_1"
                },
                "attributes" : {
                    "mlt_singlebranch" : "singlebranch",
                    "structure" : "West",
                    "welltype" : "gas injector"
                },
                "name" : "GI_1"
            },
        ]
    }
    """

    def __init__(self, ens_name: str, ens_path: Path, well_attributes_file: str):
        self._ens_name = ens_name
        self._ens_path = ens_path
        self._well_attributes_file = well_attributes_file
        self._data_raw: Dict[str, Any] = json.load(self._load_data())

        # Double Dictionary with eclipse alias as key on first level
        self._data: Dict[str, Dict[str, str]] = self._transform_data()

        # List of categories
        self._categories: List[str] = list(
            {name for categories in self._data.values() for name in categories}
        )

        # Dataframe with the well eclipse alias and well attributes as columns
        self._dataframe = (
            pd.DataFrame.from_dict(self._data, orient="index")
            .rename_axis("WELL")
            .reset_index()
        )

        # Melted dataframe with columns: WELL, CATEGORY and VALUE
        self._dataframe_melted = self._dataframe.melt(
            id_vars=["WELL"], value_vars=self._categories
        ).rename({"variable": "CATEGORY", "value": "VALUE"}, axis=1)

    def __repr__(self) -> str:
        """This is necessary for webvizstore to work on objects"""
        return f"""
WellAttributesModel({self._ens_name!r}, {self._ens_path!r}, {self._well_attributes_file!r})
        """

    @property
    def data(self) -> Dict[str, Dict[str, str]]:
        """Returns the well attributes on the transformed format"""
        return self._data

    @property
    def categories(self) -> List[str]:
        """List of all well attribute categories"""
        return self._categories

    @property
    def file_name(self) -> str:
        """Returns the well attributes file name"""
        return self._well_attributes_file

    @property
    def dataframe(self) -> pd.DataFrame:
        """Returns the well attributes data as a dataframe with the well eclipse
        alias and categories as columns
        """
        return self._dataframe

    @property
    def dataframe_melted(self) -> pd.DataFrame:
        """Returns the well attributes data as melted dataframe, that means with
        only three columns: WELL, CATEGORY and VALUE
        """
        return self._dataframe_melted

    @property
    def category_dict(self) -> Dict[str, List[str]]:
        """Returns a Dict with all possible attributes per category
        on the form:
        {
            "welltype":["oil_producer", "water_injector", ...],
            "category2": ...
        }
        """
        return {
            category: list(df["VALUE"].unique())
            for category, df in self._dataframe_melted.fillna("Undefined").groupby(
                "CATEGORY"
            )
        }

    @property
    def webviz_store(self) -> Tuple[Callable, List[Dict]]:
        return (
            self._load_data,
            [
                {
                    "self": self,
                }
            ],
        )

    @webvizstore
    def _load_data(self) -> io.BytesIO:
        """This method reads the well attributes for an ensemble. It returns
        the data from the first file it finds so it is implicitly assumed that
        the files are equal for all realizations.
        """
        ens = scratch_ensemble(self._ens_name, self._ens_path, filter_file="OK")
        df_files = ens.find_files(self._well_attributes_file)

        for _, row in df_files.iterrows():
            file_content = json.loads(Path(row["FULLPATH"]).read_text())
            logging.debug(f"Well attributes loaded from file: {row['FULLPATH']}")
            return io.BytesIO(json.dumps(file_content).encode())
        return io.BytesIO(json.dumps({"version": "0.1", "wells": []}).encode())

    def _transform_data(self) -> Dict[str, Dict[str, str]]:
        """Transforms the well attributes format to a simpler form
        where the key of the outer dictionary is the well eclipse alias:
        {
            "OP_1": {
                "mlt_singlebranch" : "mlt",
                "structure" : "East",
                "welltype" : "producer"
            },
            "OP_2 : {...}
        }
        """
        if "version" not in self._data_raw:
            raise ValueError(
                "Attribute 'version' not found in the well attributes file."
            )
        if self._data_raw["version"] != "0.1":
            raise NotImplementedError(
                f"Version {self._data_raw['version']} of the well attributes file "
                "is not implemented."
            )
        return {
            well_data["alias"]["eclipse"]: well_data["attributes"]
            for well_data in self._data_raw["wells"]
        }



from typing import Callable, Dict, List

import webviz_core_components as wcc
from dash import dcc, html

from .data_selectors_view import data_selectors_view
from .plot_options_view import plot_options_view
from .plot_traces_view import highlight_realizations_view, line_traces_view


def main_view(
    get_uuid: Callable,
    ensemble_names: List[str],
    data_column_names: List[str],
    parameter_names: List[str],
    realizations: List[int],
    initial_data: Dict,
    initial_layout: Dict,
) -> html.Div:

    return wcc.FlexBox(
        children=[
            wcc.Frame(
                style={"flex": 1, "height": "90vh"},
                children=[
                    data_stores(get_uuid=get_uuid),
                    data_selectors_view(
                        get_uuid=get_uuid,
                        ensemble_names=ensemble_names,
                        data_column_names=data_column_names,
                        parameter_names=parameter_names,
                        initial_data=initial_data,
                    ),
                    line_traces_view(
                        get_uuid=get_uuid,
                    ),
                    highlight_realizations_view(
                        get_uuid=get_uuid, realizations=realizations
                    ),
                    plot_options_view(get_uuid=get_uuid, initial_layout=initial_layout),
                ],
            ),
            wcc.Frame(
                style={"flex": 5, "height": "90vh"},
                color="white",
                highlight=False,
                children=wcc.Graph(style={"height": "86vh"}, id=get_uuid("graph")),
            ),
        ]
    )


def data_stores(get_uuid: Callable) -> html.Div:
    return html.Div(
        [
            dcc.Store(
                id={"id": get_uuid("clientside"), "plotly_attribute": "plotly_graph"},
                data=get_uuid("graph"),
            ),
            dcc.Store(
                id={"id": get_uuid("clientside"), "plotly_attribute": "plotly_data"}
            ),
            dcc.Store(
                id={
                    "id": get_uuid("clientside"),
                    "plotly_attribute": "plotly_layout",
                }
            ),
            dcc.Store(
                id={
                    "id": get_uuid("clientside"),
                    "plotly_attribute": "initial_layout",
                }
            ),
            dcc.Store(
                id={
                    "id": get_uuid("clientside"),
                    "plotly_attribute": "update_layout",
                }
            ),
            dcc.Store(id=get_uuid("stored_x_value")),
        ]
    )



from typing import Callable, Optional

import pandas as pd
import webviz_core_components as wcc
from dash import dcc
from webviz_config import WebvizConfigTheme

from webviz_subsurface._models import InplaceVolumesModel

from .comparison_layout import comparison_main_layout, comparison_selections
from .distribution_main_layout import distributions_main_layout, table_main_layout
from .filter_view import filter_layout
from .fipfile_qc_layout import (
    fipfile_qc_filters,
    fipfile_qc_main_layout,
    fipfile_qc_selections,
)
from .selections_view import selections_layout, table_selections_layout
from .tornado_view import tornado_main_layout, tornado_selections_layout


def main_view(
    get_uuid: Callable,
    volumemodel: InplaceVolumesModel,
    theme: WebvizConfigTheme,
    disjoint_set_df: Optional[pd.DataFrame] = None,
) -> dcc.Tabs:

    tabs = []
    tabs.append(
        wcc.Tab(
            label="Inplace distributions",
            value="voldist",
            children=tab_view_layout(
                main_layout=distributions_main_layout(uuid=get_uuid("main-voldist")),
                sidebar_layout=[
                    selections_layout(
                        uuid=get_uuid("selections"),
                        tab="voldist",
                        volumemodel=volumemodel,
                        theme=theme,
                    ),
                    filter_layout(
                        uuid=get_uuid("filters"), tab="voldist", volumemodel=volumemodel
                    ),
                ],
            ),
        )
    )
    tabs.append(
        wcc.Tab(
            label="Tables",
            value="table",
            children=tab_view_layout(
                main_layout=table_main_layout(
                    uuid=get_uuid("main-table"),
                ),
                sidebar_layout=[
                    table_selections_layout(
                        uuid=get_uuid("selections"),
                        tab="table",
                        volumemodel=volumemodel,
                    ),
                    filter_layout(
                        uuid=get_uuid("filters"), tab="table", volumemodel=volumemodel
                    ),
                ],
            ),
        )
    )
    if volumemodel.sensrun:
        tabs.append(
            wcc.Tab(
                label="Tornadoplots",
                value="tornado",
                children=tab_view_layout(
                    main_layout=tornado_main_layout(
                        uuid=get_uuid("main-tornado"),
                    ),
                    sidebar_layout=[
                        tornado_selections_layout(
                            uuid=get_uuid("selections"),
                            tab="tornado",
                            volumemodel=volumemodel,
                        ),
                        filter_layout(
                            uuid=get_uuid("filters"),
                            tab="tornado",
                            volumemodel=volumemodel,
                            hide_selectors=["SENSNAME_CASE"],
                        ),
                    ],
                ),
            )
        )
    if len(volumemodel.sources) > 1:
        tabs.append(
            wcc.Tab(
                label="Source comparison",
                value="src-comp",
                children=tab_view_layout(
                    main_layout=comparison_main_layout(get_uuid("main-src-comp")),
                    sidebar_layout=[
                        comparison_selections(
                            uuid=get_uuid("selections"),
                            tab="src-comp",
                            volumemodel=volumemodel,
                            compare_on="Source",
                        ),
                        filter_layout(
                            uuid=get_uuid("filters"),
                            tab="src-comp",
                            volumemodel=volumemodel,
                            hide_selectors=["SOURCE", "FLUID_ZONE"],
                        ),
                    ],
                ),
            )
        )
    if len(volumemodel.ensembles) > 1 or volumemodel.sensrun:
        compare_on = "Ensemble" if len(volumemodel.ensembles) > 1 else "Sensitivity"
        tabs.append(
            wcc.Tab(
                label=f"{compare_on} comparison",
                value="ens-comp",
                children=tab_view_layout(
                    main_layout=comparison_main_layout(get_uuid("main-ens-comp")),
                    sidebar_layout=[
                        comparison_selections(
                            uuid=get_uuid("selections"),
                            tab="ens-comp",
                            volumemodel=volumemodel,
                            compare_on=compare_on,
                        ),
                        filter_layout(
                            uuid=get_uuid("filters"),
                            tab="ens-comp",
                            volumemodel=volumemodel,
                            hide_selectors=["ENSEMBLE", "FLUID_ZONE", "SENSNAME_CASE"],
                        ),
                    ],
                ),
            )
        )
    if disjoint_set_df is not None:
        tabs.append(
            wcc.Tab(
                label="Fipfile QC",
                value="fipqc",
                children=tab_view_layout(
                    main_layout=fipfile_qc_main_layout(get_uuid("main-fipqc")),
                    sidebar_layout=[
                        fipfile_qc_selections(uuid=get_uuid("selections"), tab="fipqc"),
                        fipfile_qc_filters(
                            uuid=get_uuid("filters"),
                            tab="fipqc",
                            disjoint_set_df=disjoint_set_df,
                        ),
                    ],
                ),
            )
        )

    initial_tab = "voldist"
    if volumemodel.volume_type == "mixed":
        initial_tab = "src-comp"
    elif volumemodel.sensrun:
        initial_tab = "tornado"

    return wcc.Tabs(
        id=get_uuid("tabs"),
        value=initial_tab,
        style={"width": "100%"},
        persistence=True,
        children=tabs,
    )


def tab_view_layout(main_layout: list, sidebar_layout: list) -> wcc.FlexBox:
    return wcc.FlexBox(
        children=[
            wcc.Frame(
                style={"flex": 1, "height": "87vh"},
                children=sidebar_layout,
            ),
            wcc.Frame(
                style={"flex": 6, "height": "87vh"},
                color="white",
                highlight=False,
                children=main_layout,
            ),
        ]
    )



from .ensemble_table_provider import ColumnMetadata, EnsembleTableProvider
from .ensemble_table_provider_factory import EnsembleTableProviderFactory
from .ensemble_table_provider_impl_arrow import EnsembleTableProviderImplArrow



import datetime
from typing import List, Optional, Sequence, Tuple

import pandas as pd
from webviz_subsurface_components import ExpressionInfo

from webviz_subsurface._providers import EnsembleSummaryProvider, Frequency
from webviz_subsurface._utils.dataframe_utils import make_date_column_datetime_object
from webviz_subsurface._utils.vector_calculator import (
    create_calculated_vector_df,
    get_selected_expressions,
)

from .. import dataframe_utils
from ..from_timeseries_cumulatives import (
    calculate_from_resampled_cumulative_vectors_df,
    get_cumulative_vector_name,
    is_per_interval_or_per_day_vector,
)
from .derived_vectors_accessor import DerivedVectorsAccessor


class DerivedDeltaEnsembleVectorsAccessorImpl(DerivedVectorsAccessor):
    """
    Class to create derived vector data and access these for a delta ensemble.

    The delta ensemble is represented a pair of two ensemble summary providers.

    A list of vector names are provided, and data is fetched or created based on which
    type of vectors are present in the list.

    Vector names can be regular vectors existing among vector names in the providers, Per
    Interval/Per Day vector or a calculated vector from vector calculator.

    Based on the vector type, the class provides an interface for retrieving dataframes
    for the set of such vectors for the provider.
    """

    def __init__(
        self,
        name: str,
        provider_pair: Tuple[EnsembleSummaryProvider, EnsembleSummaryProvider],
        vectors: List[str],
        expressions: Optional[List[ExpressionInfo]] = None,
        resampling_frequency: Optional[Frequency] = None,
        relative_date: Optional[datetime.datetime] = None,
    ) -> None:
        if len(provider_pair) != 2:
            raise ValueError(
                'Expect input argument "provider_pair" to have two providers!'
                f"Got {len(provider_pair)}"
            )
        self._provider_a = provider_pair[0]
        self._provider_b = provider_pair[1]

        # Initialize base class
        _intersected_realizations = [
            elm
            for elm in self._provider_a.realizations()
            if elm in self._provider_b.realizations()
        ]
        super().__init__(_intersected_realizations)

        self._name = name
        if (
            self._provider_a.supports_resampling()
            != self._provider_b.supports_resampling()
        ):
            raise ValueError(
                f"Ensemble A and B must have same resampling support! "
                f"Ensemble A support resampling: {self._provider_a.supports_resampling()} "
                f"and Ensemble B support resampling: {self._provider_b.supports_resampling()}"
            )

        # Intersection of vectors in providers
        _accessor_vectors = [
            elm
            for elm in self._provider_a.vector_names()
            if elm in self._provider_b.vector_names()
        ]

        # Categorize vector types among the vectors in argument
        self._provider_vectors = [
            vector for vector in vectors if vector in _accessor_vectors
        ]
        self._per_interval_and_per_day_vectors = [
            vector
            for vector in vectors
            if is_per_interval_or_per_day_vector(vector)
            and get_cumulative_vector_name(vector) in _accessor_vectors
        ]
        self._vector_calculator_expressions = (
            get_selected_expressions(expressions, vectors)
            if expressions is not None
            else []
        )

        # Set resampling frequency
        self._resampling_frequency = (
            resampling_frequency
            if self._provider_a.supports_resampling()
            and self._provider_b.supports_resampling()
            else None
        )

        self._relative_date = relative_date

    def __create_delta_ensemble_vectors_df(
        self,
        vector_names: List[str],
        resampling_frequency: Optional[Frequency],
        realizations: Optional[Sequence[int]] = None,
    ) -> pd.DataFrame:
        """
        Get vectors dataframe with delta vectors for ensemble A and B, for common realizations

        `Return:` Dataframe with delta ensemble data for common vectors and realizations in ensemble
        A and B.

        `Output:`
        * DataFrame with columns ["DATE", "REAL", vector1, ..., vectorN]

        `Input:`
        * vector_names: List[str] - List of vector names to get data for
        * resampling_frequency: Optional[Frequency] - Optional resampling frequency
        * realizations: Optional[Sequence[int]] - Optional sequence of realization numbers for
        vectors

        NOTE:
        - Performs "inner join". Only obtain matching index ["DATE", "REAL"] - i.e "DATE"-"REAL"
        combination present in only one vector -> neglected
        - Ensures equal dates samples and realizations by dropping nan-values
        """

        if not vector_names:
            raise ValueError("List of requested vector names is empty")

        # NOTE: index order ["DATE","REAL"] to obtain column order when
        # performing reset_index() later
        ensemble_a_vectors_df = self._provider_a.get_vectors_df(
            vector_names, resampling_frequency, realizations
        ).set_index(["DATE", "REAL"])
        ensemble_b_vectors_df = self._provider_b.get_vectors_df(
            vector_names, resampling_frequency, realizations
        ).set_index(["DATE", "REAL"])

        # Reset index, sort values by "REAL" and thereafter by "DATE" to
        # group realizations and order by date
        ensembles_delta_vectors_df = (
            ensemble_a_vectors_df.sub(ensemble_b_vectors_df)
            .dropna(axis=0, how="any")
            .reset_index()
            .sort_values(["REAL", "DATE"], ignore_index=True)
        )

        make_date_column_datetime_object(ensembles_delta_vectors_df)

        return ensembles_delta_vectors_df

    def has_provider_vectors(self) -> bool:
        return len(self._provider_vectors) > 0

    def has_per_interval_and_per_day_vectors(self) -> bool:
        return len(self._per_interval_and_per_day_vectors) > 0

    def has_vector_calculator_expressions(self) -> bool:
        return len(self._vector_calculator_expressions) > 0

    def get_provider_vectors_df(
        self, realizations: Optional[Sequence[int]] = None
    ) -> pd.DataFrame:
        """
        Get vectors dataframe with delta vectors for ensemble A and B, for common realizations and
        selected vectors

        `Return:` Dataframe with delta ensemble data for common vectors and realizations in ensemble
        A and B.

        `Output:`
        * DataFrame with columns ["DATE", "REAL", vector1, ..., vectorN]
        """
        if not self.has_provider_vectors():
            raise ValueError(
                f'Vector data handler for provider "{self._name}" has no provider vectors'
            )

        if self._relative_date:
            return dataframe_utils.create_relative_to_date_df(
                self.__create_delta_ensemble_vectors_df(
                    self._provider_vectors, self._resampling_frequency, realizations
                ),
                self._relative_date,
            )
        return self.__create_delta_ensemble_vectors_df(
            self._provider_vectors, self._resampling_frequency, realizations
        )

    def create_per_interval_and_per_day_vectors_df(
        self,
        realizations: Optional[Sequence[int]] = None,
    ) -> pd.DataFrame:
        """Get dataframe with interval delta and per day delta vector data for provided vectors.

        The returned dataframe contains columns with name of vector and corresponding interval delta
        or per day delta data.

        Interval delta and per day delta date is calculated with same sampling frequency as provider
        is set with. I.e. resampling frequency is given for providers supporting resampling,
        otherwise sampling frequency is fixed.

        `Input:`
        * realizations: Sequence[int] - Sequency of realization numbers to include in calculation

        `Output:`
        * dataframe with interval  vector names in columns and their cumulative data in rows.
        `Columns` in dataframe: ["DATE", "REAL", vector1, ..., vectorN]

        ---------------------
        `NOTE:`
        * Handle calculation of cumulative when raw data is added
        * See TODO in calculate_from_resampled_cumulative_vectors_df()
        """
        if not self.has_per_interval_and_per_day_vectors():
            raise ValueError(
                f'Vector data handler for provider "{self._name}" has no per interval '
                "or per day vector names"
            )

        cumulative_vector_names = [
            get_cumulative_vector_name(elm)
            for elm in self._per_interval_and_per_day_vectors
            if is_per_interval_or_per_day_vector(elm)
        ]
        cumulative_vector_names = list(sorted(set(cumulative_vector_names)))

        vectors_df = self.__create_delta_ensemble_vectors_df(
            cumulative_vector_names, self._resampling_frequency, realizations
        )

        per_interval_and_per_day_vectors_df = pd.DataFrame()
        for vector_name in self._per_interval_and_per_day_vectors:
            cumulative_vector_name = get_cumulative_vector_name(vector_name)
            per_interval_or_per_day_vector_df = (
                calculate_from_resampled_cumulative_vectors_df(
                    vectors_df[["DATE", "REAL", cumulative_vector_name]],
                    as_per_day=vector_name.startswith("PER_DAY_"),
                )
            )
            if per_interval_and_per_day_vectors_df.empty:
                per_interval_and_per_day_vectors_df = per_interval_or_per_day_vector_df
            else:
                per_interval_and_per_day_vectors_df = pd.merge(
                    per_interval_and_per_day_vectors_df,
                    per_interval_or_per_day_vector_df,
                    how="inner",
                )

        if self._relative_date:
            return dataframe_utils.create_relative_to_date_df(
                per_interval_and_per_day_vectors_df,
                self._relative_date,
            )
        return per_interval_and_per_day_vectors_df

    def create_calculated_vectors_df(
        self, realizations: Optional[Sequence[int]] = None
    ) -> pd.DataFrame:
        """Get dataframe with calculated vector data for provided vectors.

        The returned dataframe contains columns with name of vector and corresponding calculated
        data.

        The calculated vectors for delta ensembles are created by first creating the calculated
        vector data for ensemble A and B separately, and thereafter subtracting the data in ensemble
        B from A. Thereby one obtain creating the delta ensemble of the resulting calculated
        vectors.

        Calculated vectors are created with same sampling frequency as providers is set with. I.e.
        resampling frequency is given for providers supporting resampling, otherwise sampling
        frequency is fixed.

        `Input:`
        * realizations: Sequence[int] - Sequency of realization numbers to include in calculation

        `Output:`
        * dataframe with vector names in columns and their calculated data in rows
        `Columns` in dataframe: ["DATE", "REAL", vector1, ..., vectorN]
        """
        if not self.has_vector_calculator_expressions():
            raise ValueError(
                f'Assembled vector data accessor for provider "{self._name}"'
                "has no vector calculator expressions"
            )

        provider_a_calculated_vectors_df = pd.DataFrame()
        provider_b_calculated_vectors_df = pd.DataFrame()
        for expression in self._vector_calculator_expressions:
            provider_a_calculated_vector_df = create_calculated_vector_df(
                expression, self._provider_a, realizations, self._resampling_frequency
            )
            provider_b_calculated_vector_df = create_calculated_vector_df(
                expression, self._provider_b, realizations, self._resampling_frequency
            )

            if (
                provider_a_calculated_vector_df.empty
                or provider_b_calculated_vector_df.empty
            ):
                # TODO: Consider raising ValueError of vector calculation in one provider fails?
                # If both fails, it's okay?
                continue

            def __inner_merge_dataframes(
                first: pd.DataFrame, second: pd.DataFrame
            ) -> pd.DataFrame:
                if first.empty:
                    return second
                return pd.merge(first, second, how="inner")

            provider_a_calculated_vectors_df = __inner_merge_dataframes(
                provider_a_calculated_vectors_df, provider_a_calculated_vector_df
            )
            provider_b_calculated_vectors_df = __inner_merge_dataframes(
                provider_b_calculated_vectors_df, provider_b_calculated_vector_df
            )

        # NOTE: index order ["DATE","REAL"] to obtain column order when
        # performing reset_index() later
        provider_a_calculated_vectors_df.set_index(["DATE", "REAL"], inplace=True)
        provider_b_calculated_vectors_df.set_index(["DATE", "REAL"], inplace=True)

        # Reset index, sort values by "REAL" and thereafter by "DATE" to
        # group realizations and order by date
        delta_ensemble_calculated_vectors_df = (
            provider_a_calculated_vectors_df.sub(provider_b_calculated_vectors_df)
            .dropna(axis=0, how="any")
            .reset_index()
            .sort_values(["REAL", "DATE"], ignore_index=True)
        )

        make_date_column_datetime_object(delta_ensemble_calculated_vectors_df)

        if self._relative_date:
            return dataframe_utils.create_relative_to_date_df(
                delta_ensemble_calculated_vectors_df,
                self._relative_date,
            )
        return delta_ensemble_calculated_vectors_df



import logging
from typing import Dict, List, Optional
from urllib.parse import quote

import flask
import geojson
from dash import Dash

from webviz_subsurface._providers.well_provider.well_provider import WellProvider
from webviz_subsurface._utils.perf_timer import PerfTimer

LOGGER = logging.getLogger(__name__)

_ROOT_URL_PATH = "/WellServer"

_WELL_SERVER_INSTANCE: Optional["WellServer"] = None


class WellServer:
    def __init__(self, app: Dash) -> None:
        self._setup_url_rule(app)
        self._id_to_provider_dict: Dict[str, WellProvider] = {}

    @staticmethod
    def instance(app: Dash) -> "WellServer":
        # pylint: disable=global-statement
        global _WELL_SERVER_INSTANCE
        if not _WELL_SERVER_INSTANCE:
            LOGGER.debug("Initializing WellServer instance")
            _WELL_SERVER_INSTANCE = WellServer(app)

        return _WELL_SERVER_INSTANCE

    def add_provider(self, provider: WellProvider) -> None:

        provider_id = provider.provider_id()
        LOGGER.debug(f"Adding provider with id={provider_id}")

        existing_provider = self._id_to_provider_dict.get(provider_id)
        if existing_provider:
            # Issue a warning if there already is a provider registered with the same
            # id AND if the actual provider instance is different.
            # This should not be a problem, but will happen until the provider factory
            # gets caching.
            if existing_provider is not provider:
                LOGGER.warning(
                    f"Provider with id={provider_id} ignored, the id is already present"
                )
                return

        self._id_to_provider_dict[provider_id] = provider

    def encode_partial_url(
        self,
        provider_id: str,
        well_names: List[str],
    ) -> str:

        if not provider_id in self._id_to_provider_dict:
            raise ValueError("Could not find provider")

        sorted_well_names_str = "~".join(sorted(well_names))

        url_path: str = (
            f"{_ROOT_URL_PATH}/{quote(provider_id)}/{quote(sorted_well_names_str)}"
        )

        return url_path

    def _setup_url_rule(self, app: Dash) -> None:
        @app.server.route(_ROOT_URL_PATH + "/<provider_id>/<well_names_str>")
        def _handle_wells_request(
            provider_id: str, well_names_str: str
        ) -> flask.Response:
            LOGGER.debug(
                f"Handling well request: "
                f"provider_id={provider_id} "
                f"well_names_str={well_names_str} "
            )

            timer = PerfTimer()

            try:
                provider = self._id_to_provider_dict[provider_id]
                well_names_arr = well_names_str.split("~")
            # pylint: disable=bare-except
            except:
                LOGGER.error("Error decoding wells address")
                flask.abort(404)

            validate_geometry = True
            feature_arr = []
            for wname in well_names_arr:
                well_path = provider.get_well_path(wname)

                coords = list(zip(well_path.x_arr, well_path.y_arr, well_path.z_arr))
                # coords = coords[0::20]
                point = geojson.Point(
                    coordinates=[coords[0][0], coords[0][1]], validate=validate_geometry
                )

                geocoll = geojson.GeometryCollection(geometries=[point])

                feature = geojson.Feature(
                    id=wname, geometry=geocoll, properties={"name": wname}
                )
                feature_arr.append(feature)

            featurecoll = geojson.FeatureCollection(features=feature_arr)
            response = flask.Response(
                geojson.dumps(featurecoll), mimetype="application/geo+json"
            )

            LOGGER.debug(f"Request handled in: {timer.elapsed_s():.2f}s")
            return response



from typing import Callable, Optional, Union

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from dash import Input, Output, State, callback, callback_context, dash_table, html
from dash.exceptions import PreventUpdate

from webviz_subsurface._figures import create_figure
from webviz_subsurface._models import InplaceVolumesModel

from ..utils.table_and_figure_utils import (
    add_correlation_line,
    create_data_table,
    create_table_columns,
)
from ..utils.utils import move_to_end_of_list
from ..views.comparison_layout import (
    comparison_qc_plots_layout,
    comparison_table_layout,
)


# pylint: disable=too-many-locals
def comparison_controllers(
    get_uuid: Callable,
    volumemodel: InplaceVolumesModel,
) -> None:
    @callback(
        Output({"id": get_uuid("main-src-comp"), "wrapper": "table"}, "children"),
        Input(get_uuid("selections"), "data"),
        Input({"id": get_uuid("main-src-comp"), "element": "display-option"}, "value"),
        State(get_uuid("page-selected"), "data"),
    )
    def _update_page_src_comp(
        selections: dict,
        display_option: str,
        page_selected: str,
    ) -> html.Div:
        ctx = callback_context.triggered[0]

        if page_selected != "src-comp":
            raise PreventUpdate

        selections = selections[page_selected]
        if not "display-option" in ctx["prop_id"]:
            if not selections["update"]:
                raise PreventUpdate

        return comparison_callback(
            compare_on="SOURCE",
            volumemodel=volumemodel,
            selections=selections,
            display_option=display_option,
        )

    @callback(
        Output({"id": get_uuid("main-ens-comp"), "wrapper": "table"}, "children"),
        Input(get_uuid("selections"), "data"),
        Input({"id": get_uuid("main-ens-comp"), "element": "display-option"}, "value"),
        State(get_uuid("page-selected"), "data"),
    )
    def _update_page_ens_comp(
        selections: dict,
        display_option: str,
        page_selected: str,
    ) -> html.Div:
        ctx = callback_context.triggered[0]

        if page_selected != "ens-comp":
            raise PreventUpdate

        selections = selections[page_selected]
        if not "display-option" in ctx["prop_id"]:
            if not selections["update"]:
                raise PreventUpdate

        return comparison_callback(
            compare_on="SENSNAME_CASE"
            if selections["compare_on"] == "Sensitivity"
            else "ENSEMBLE",
            volumemodel=volumemodel,
            selections=selections,
            display_option=display_option,
        )


def comparison_callback(
    compare_on: str,
    volumemodel: InplaceVolumesModel,
    selections: dict,
    display_option: str,
) -> html.Div:
    if selections["value1"] == selections["value2"]:
        return html.Div("Comparison between equal data")

    # Handle None in highlight criteria input
    for key in ["Accept value", "Ignore <"]:
        selections[key] = selections[key] if selections[key] is not None else 0

    groupby = selections["Group by"] if selections["Group by"] is not None else []
    group_on_fluid = "FLUID_ZONE" in groupby
    # for hc responses and bo/bg the data should be grouped
    # on fluid zone to avoid misinterpretations
    if (
        selections["Response"] in volumemodel.hc_responses + ["BO", "BG"]
        and "FLUID_ZONE" not in groupby
    ):
        groupby.append("FLUID_ZONE")

    if display_option == "multi-response table":
        # select max one hc_response for a cleaner table
        responses = [selections["Response"]] + [
            col
            for col in volumemodel.responses
            if col not in volumemodel.hc_responses and col != selections["Response"]
        ]
        df = create_comparison_df(
            volumemodel,
            compare_on=compare_on,
            selections=selections,
            responses=responses,
            abssort_on=f"{selections['Response']} diff (%)",
            groups=groupby,
        )
        if df.empty:
            return html.Div("No data left after filtering")

        return comparison_table_layout(
            table=create_comaprison_table(
                tabletype=display_option,
                df=df,
                groupby=groupby,
                selections=selections,
                compare_on=compare_on,
                volumemodel=volumemodel,
            ),
            table_type=display_option,
            selections=selections,
            filter_info="SOURCE" if compare_on != "SOURCE" else "ENSEMBLE",
        )

    if compare_on == "SOURCE" or "REAL" in groupby:
        diffdf_real = create_comparison_df(
            volumemodel,
            compare_on=compare_on,
            selections=selections,
            responses=[selections["Response"]],
            groups=groupby + (["REAL"] if "REAL" not in groupby else []),
            rename_diff_col=True,
        )

    if "REAL" not in groupby:
        diffdf_group = create_comparison_df(
            volumemodel,
            compare_on=compare_on,
            selections=selections,
            responses=[selections["Response"]],
            groups=groupby,
            rename_diff_col=True,
        )
        if compare_on == "SOURCE" and not diffdf_group.empty:
            # Add column with number of highlighted realizations
            diffdf_group["💡 reals"] = diffdf_group.apply(
                lambda row: find_higlighted_real_count(row, diffdf_real, groupby),
                axis=1,
            )

    df = diffdf_group if "REAL" not in groupby else diffdf_real
    if df.empty:
        return html.Div("No data left after filtering")

    if display_option == "single-response table":
        return comparison_table_layout(
            table=create_comaprison_table(
                tabletype=display_option,
                df=df,
                groupby=groupby,
                selections=selections,
                use_si_format=selections["Response"] in volumemodel.volume_columns,
                compare_on=compare_on,
            ),
            table_type=display_option,
            selections=selections,
            filter_info="SOURCE" if compare_on != "SOURCE" else "ENSEMBLE",
        )

    if display_option == "plots":
        if "|" in selections["value1"]:
            ens1, sens1 = selections["value1"].split("|")
            ens2, sens2 = selections["value2"].split("|")
            value1, value2 = (sens1, sens2) if ens1 == ens2 else (ens1, ens2)
        else:
            value1, value2 = selections["value1"], selections["value2"]

        resp1 = f"{selections['Response']} {value1}"
        resp2 = f"{selections['Response']} {value2}"

        scatter_corr = create_scatterfig(
            df=df, x=resp1, y=resp2, selections=selections, groupby=groupby
        )
        scatter_corr = add_correlation_line(
            figure=scatter_corr, xy_min=df[resp1].min(), xy_max=df[resp1].max()
        )
        scatter_diff_vs_response = create_scatterfig(
            df=df,
            x=resp1,
            y=selections["Diff mode"],
            selections=selections,
            groupby=groupby,
            diff_mode=selections["Diff mode"],
        )
        scatter_diff_vs_real = (
            create_scatterfig(
                df=diffdf_real,
                x="REAL",
                y=selections["Diff mode"],
                selections=selections,
                groupby=groupby,
                diff_mode=selections["Diff mode"],
            )
            if compare_on == "SOURCE"
            else None
        )
        barfig_non_highlighted = create_barfig(
            df=df[df["highlighted"] == "yes"],
            groupby=groupby
            if group_on_fluid
            else [x for x in groupby if x != "FLUID_ZONE"],
            diff_mode=selections["Diff mode"],
            colorcol=resp1,
        )

    return comparison_qc_plots_layout(
        scatter_diff_vs_real,
        scatter_corr,
        scatter_diff_vs_response,
        barfig_non_highlighted,
    )


def create_comparison_df(
    volumemodel: InplaceVolumesModel,
    compare_on: str,
    responses: list,
    selections: dict,
    groups: list,
    abssort_on: str = "diff (%)",
    rename_diff_col: bool = False,
) -> pd.DataFrame:

    resp = selections["Response"]
    adiitional_groups = [
        x for x in ["SOURCE", "ENSEMBLE", "SENSNAME_CASE"] if x in volumemodel.selectors
    ]
    groups = groups + adiitional_groups
    df = volumemodel.get_df(selections["filters"], groups=groups)

    # filter dataframe and set values to compare against
    if not "|" in selections["value1"]:
        value1, value2 = selections["value1"], selections["value2"]
        df = df[df[compare_on].isin([value1, value2])]
    else:
        ens1, sens1 = selections["value1"].split("|")
        ens2, sens2 = selections["value2"].split("|")
        if ens1 == ens2:
            compare_on = "SENSNAME_CASE"
        value1, value2 = (sens1, sens2) if ens1 == ens2 else (ens1, ens2)

        df = df[
            ((df["ENSEMBLE"] == ens1) & (df["SENSNAME_CASE"] == sens1))
            | ((df["ENSEMBLE"] == ens2) & (df["SENSNAME_CASE"] == sens2))
        ]

    # if no data left, or one of the selected SOURCE/ENSEMBLE is not present
    # in the dataframe after filtering, return empty dataframe
    if df.empty or any(x not in df[compare_on].values for x in [value1, value2]):
        return pd.DataFrame()

    df = df.loc[:, groups + responses].pivot_table(
        columns=compare_on,
        index=[x for x in groups if x not in [compare_on, "SENSNAME_CASE"]],
    )
    responses = [x for x in responses if x in df]
    for col in responses:
        df[col, "diff"] = df[col][value2] - df[col][value1]
        df[col, "diff (%)"] = ((df[col][value2] / df[col][value1]) - 1) * 100
        df.loc[df[col]["diff"] == 0, (col, "diff (%)")] = 0
    df = df[responses].replace([np.inf, -np.inf], np.nan).reset_index()

    # remove rows where the selected response is nan
    # can happen for properties where the volume columns are 0
    df = df.loc[~((df[resp][value1].isna()) & (df[resp][value2].isna()))]
    if selections["Remove zeros"]:
        df = df.loc[~((df[resp]["diff"] == 0) & (df[resp][value1] == 0))]

    df["highlighted"] = compute_highlighted_col(df, resp, value1, selections)
    df.columns = df.columns.map(" ".join).str.strip(" ")

    # remove BO∕BG columns if they are nan and drop SOURCE/ENSMEBLE column
    dropcols = [
        x for x in df.columns[df.isna().all()] if x.split(" ")[0] in ["BO", "BG"]
    ] + adiitional_groups
    df = df[[x for x in df.columns if x not in dropcols]]

    if rename_diff_col:
        df = df.rename(columns={f"{resp} diff": "diff", f"{resp} diff (%)": "diff (%)"})
    df = add_fluid_zone_column(df, selections["filters"])
    return df.sort_values(by=[abssort_on], key=abs, ascending=False)


def compute_highlighted_col(
    df: pd.DataFrame, response: str, value1: str, selections: dict
) -> np.ndarray:
    highlight_mask = (df[response][value1] > selections["Ignore <"]) & (
        df[response]["diff (%)"].abs() > selections["Accept value"]
    )
    return np.where(highlight_mask, "yes", "no")


def find_higlighted_real_count(
    row: pd.Series, df_per_real: pd.DataFrame, groups: list
) -> str:
    query = " & ".join([f"{col}=='{row[col]}'" for col in groups])
    result = df_per_real.query(query) if groups else df_per_real
    return str(len(result[result["highlighted"] == "yes"]))


def create_comaprison_table(
    tabletype: str,
    df: pd.DataFrame,
    groupby: list,
    selections: dict,
    compare_on: str,
    use_si_format: Optional[bool] = None,
    volumemodel: Optional[InplaceVolumesModel] = None,
) -> dash_table.DataTable:

    diff_mode_percent = selections["Diff mode"] == "diff (%)"

    if selections["Remove non-highlighted"]:
        df = df.loc[df["highlighted"] == "yes"]
        if df.empty:
            return html.Div(
                [
                    html.Div("All data outside highlight criteria!"),
                    html.Div(
                        "To see the data turn off setting 'Display only highlighted data'"
                    ),
                ]
            )

    if tabletype == "multi-response table":
        diff_cols = [x for x in df.columns if x.endswith(selections["Diff mode"])]
        rename_dict = {x: x.split(" ")[0] for x in diff_cols}
        df = df[groupby + diff_cols + ["highlighted"]].rename(columns=rename_dict)
        df = add_fluid_zone_column(df, selections["filters"])

        columns = create_table_columns(
            columns=move_to_end_of_list("FLUID_ZONE", df.columns),
            text_columns=groupby,
            use_si_format=volumemodel.volume_columns
            if volumemodel is not None and not diff_mode_percent
            else None,
            use_percentage=list(df.columns) if diff_mode_percent else None,
        )
    else:
        columns = create_table_columns(
            columns=move_to_end_of_list("FLUID_ZONE", df.columns),
            text_columns=groupby,
            use_si_format=list(df.columns) if use_si_format else None,
            use_percentage=["diff (%)"],
        )

    return create_data_table(
        selectors=groupby,
        columns=columns,
        height="80vh",
        data=df.to_dict("records"),
        table_id={"table_id": f"{compare_on}-comp-table"},
        style_cell={"textAlign": "center"},
        style_data_conditional=[
            {
                "if": {"filter_query": "{highlighted} = 'yes'"},
                "backgroundColor": "rgb(230, 230, 230)",
                "fontWeight": "bold",
            },
        ],
        style_cell_conditional=[
            {"if": {"column_id": "highlighted"}, "display": "None"}
        ],
    )


def create_scatterfig(
    df: pd.DataFrame,
    x: str,
    y: str,
    selections: dict,
    groupby: list,
    diff_mode: Optional[str] = None,
) -> go.Figure:

    highlight_colors = {"yes": "#FF1243", "no": "#80B7BC"}
    colorby = (
        selections["Color by"]
        if selections["Color by"] == "highlighted"
        else groupby[0]
    )
    df[colorby] = df[colorby].astype(str)

    fig = (
        create_figure(
            plot_type="scatter",
            data_frame=df,
            x=x,
            y=y,
            color_discrete_sequence=px.colors.qualitative.Dark2,
            color_discrete_map=highlight_colors if colorby == "highlighted" else None,
            color=colorby,
            hover_data=groupby,
        )
        .update_traces(marker_size=10)
        .update_layout(margin={"l": 20, "r": 20, "t": 20, "b": 20})
    )
    if len(df) == 1:
        fig.update_xaxes(range=[df[x].mean() * 0.95, df[x].mean() * 1.05])
    if diff_mode is not None:
        fig.update_yaxes(range=find_diff_plot_range(df, diff_mode, selections))
        if diff_mode == "diff (%)" and y == diff_mode:
            fig.add_hline(y=selections["Accept value"], line_dash="dot").add_hline(
                y=-selections["Accept value"], line_dash="dot"
            )
    return fig


def find_diff_plot_range(df: pd.DataFrame, diff_mode: str, selections: dict) -> list:
    """
    Find plot range for diff axis. If axis focus is selected
    the range will center around the non-acepted data points.
    An 10% extension is added to the axis
    """
    if selections["Axis focus"] and "yes" in df["highlighted"].values:
        df = df[df["highlighted"] == "yes"]

    low = min(df[diff_mode].min(), -selections["Accept value"])
    high = max(df[diff_mode].max(), selections["Accept value"])
    extend = (high - low) * 0.1
    return [low - extend, high + extend]


def create_barfig(
    df: pd.DataFrame, groupby: list, diff_mode: str, colorcol: str
) -> Union[None, go.Figure]:
    if df.empty:
        return None
    return (
        create_figure(
            plot_type="bar",
            data_frame=df,
            x=df[groupby].astype(str).agg(" ".join, axis=1) if groupby else ["Total"],
            y=diff_mode,
            color_continuous_scale="teal_r",
            color=df[colorcol],
            hover_data={col: True for col in groupby},
            opacity=1,
        )
        .update_layout(
            margin={"l": 20, "r": 20, "t": 5, "b": 5},
            bargap=0.15,
            paper_bgcolor="rgba(0,0,0,0)",
        )
        .update_xaxes(title_text=None, tickangle=45, ticks="outside")
        .update_yaxes(zeroline=True, zerolinecolor="black")
    )


def add_fluid_zone_column(dframe: pd.DataFrame, filters: dict) -> pd.DataFrame:
    if "FLUID_ZONE" not in dframe and "FLUID_ZONE" in filters:
        dframe["FLUID_ZONE"] = (" + ").join(filters["FLUID_ZONE"])
    return dframe



########################################
#
#  Copyright (C) 2021-     Equinor ASA
#
#  webviz-subsurface is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  webviz-subsurface is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http:#www.gnu.org/licenses/gpl.html>
#  for more details.
#
########################################
########################################
#
#  The code in this file is based on and/or inspired by opm-common,
#  which is distributed under the GNU General Public License v3.0,
#  and available at https://github.com/OPM/opm-common.
#
#  Especially in the following points this file, or files imported in this file
#  within the same context, make use of or are based on methods developed
#  for the opm-common project:
#  - Reading of Eclipse INIT files
#  - Units, unit systems and conversion of units
#  - Class structures for storing and accessing PVT data,
#    both base classes and derived classes for specific fluids
#
#  You should have received a copy of the GNU General Public License
#  along with webviz-subsurface. If not, see (https://www.gnu.org/licenses/gpl-3.0.html).
#
#  (12-01-2021)
#
########################################

from typing import Any, Callable, List, Optional, Union

import numpy as np

# opm is only available for Linux,
# hence, ignore any import exception here to make
# it still possible to use the PvtPlugin on
# machines with other OSes.
#
# NOTE: Functions in this file cannot be used
#       on non-Linux OSes.
try:
    from opm.io.ecl import EclFile
except ImportError:
    pass

from ..eclipse_unit import ConvertUnits, CreateUnitConverter, EclUnitEnum, EclUnits
from .pvt_common import (
    EclPhaseIndex,
    EclPropertyTableRawData,
    FluidImplementation,
    InitFileDefinitions,
    PvxOBase,
    surface_mass_density,
)


class WaterImpl(PvxOBase):
    """Class holding a PVT interpolant and access methods for PVT data for water"""

    def __init__(
        self,
        index_table: int,
        raw: EclPropertyTableRawData,
        surface_mass_density_water: float,
        convert: ConvertUnits,
    ) -> None:
        """Initializes a Water object.

        Creates an interpolant for Water for the Eclipse data given
        by the raw data in the table with the given index.

        Args:
            index_table: Index of the PVT table
            raw: Eclipse raw data
            surface_mass_density_water: Surface mass density of water
            convert: Tuple holding a callable and a ConvertUnits object for unit conversions

        """
        column_stride = raw.num_rows * raw.num_tables * raw.num_primary
        table_stride = raw.num_primary * raw.num_rows
        current_stride = index_table * table_stride

        # [ Pw, 1/Bw, Cw, 1/(Bw*mu_w), Cw - Cv ]
        self.__pw_ref = convert.independent(raw.data[current_stride])
        self.__recip_fvf_ref = convert.column[0](
            raw.data[current_stride + 1 * column_stride]
        )
        self.__c_w_ref = convert.column[1](raw.data[current_stride + 2 * column_stride])
        self.__recip_fvf_visc_ref = convert.column[2](
            raw.data[current_stride + 3 * column_stride]
        )
        self.__diff_cw_cv_ref = convert.column[3](
            raw.data[current_stride + 4 * column_stride]
        )

        self.__surface_mass_density_water = surface_mass_density_water

    def __recip_fvf(self, p_w: float) -> float:
        """Computes the reciprocal of the formation volume factor for the given water pressure.

        Args:
            p_w: Water pressure

        Returns:
            Reciprocal of the formation volume factor

        """

        # B_w(P) = (B_w(P_ref)) / (1 + x + (x^2 / 2))
        # x = C * (P - P_ref)
        # NOTE: Don't forget that all values in INIT files are stored as reciprocals!
        # 1 / B_w(P) = 1 / B_w(P_ref) * (1 + x + x^2 / 2)

        x = self.__c_w_ref * (p_w - self.__pw_ref)

        return self.__recip_fvf_ref * self.__exp(x)

    def __recip_fvf_visc(self, p_w: float) -> float:
        """Computes the reciprocal of the product of formation volume factor
        and viscosity for the given water pressure.

        Args:
            p_w: Water pressure

        Returns:
            Reciprocal of the product of formation volume factor and viscosity

        """

        # mu_w(P) * B_w(P) = B_w(P_ref) * mu_w(P_ref) / (1 + y + (y^2 / 2))
        # y = (C - C_v) * (P - P_ref)
        # NOTE: Don't forget that all values in INIT files are stored as reciprocals!
        # 1 / (mu_w(P) * B_w(P)) = 1 / (B_w(P_ref) * mu_w(P_ref)) * (1 + y + (y^2 / 2))

        y = self.__diff_cw_cv_ref * (p_w - self.__pw_ref)

        return self.__recip_fvf_visc_ref * self.__exp(y)

    @staticmethod
    def __evaluate(
        pressures: np.ndarray, calculate: Callable[[Any], Any]
    ) -> np.ndarray:
        """Calls the calculate method with each of the pressure values
        and returns the results.

        Args:
            pressures: Pressure values
            calculate: Method to be called with each of the pressure values

        Returns:
            Result values

        """
        return np.array([calculate(pressure) for pressure in pressures])

    @staticmethod
    def __exp(x: float) -> float:
        """Internal helper function.

        Computes an approximation to the exponential function by means
        of a Taylor series (to second degree).

        Args:
            x: Expression to use (see manual)

        Returns:
            Approximation to exponential function

        """
        return 1.0 + x * (1.0 + x / 2.0)

    def formation_volume_factor(
        self, ratio: np.ndarray, pressure: np.ndarray
    ) -> np.ndarray:
        """Computes formation volume factor values
        for the given pressure values.

        Args:
            ratio: Dummy argument, only to conform to interface of base class
            pressure: Pressure values

        Returns:
            Formation volume factor values for the given pressure values.

        """
        return self.__evaluate(pressure, lambda p: 1.0 / self.__recip_fvf(p))

    def viscosity(self, ratio: np.ndarray, pressure: np.ndarray) -> np.ndarray:
        """Computes viscosity values for the given pressure values.

        Args:
            ratio: Dummy argument, only to conform to interface of base class
            pressure: Pressure values

        Returns:
            Viscosity values for the given pressure values.

        """
        return self.__evaluate(
            pressure, lambda p: self.__recip_fvf(p) / self.__recip_fvf_visc(p)
        )

    def density(self, ratio: np.ndarray, pressure: np.ndarray) -> np.ndarray:
        """Args:
            ratio: Dummy argument, only to conform to interface of base class.
            pressure: Pressure values the density values are requested for.

        Returns:
            Density values for the given ratio and pressure values.

        """
        # rho_w = rho_w,sc / B_w
        fvf_water = self.formation_volume_factor(ratio, pressure)
        return np.array(
            [
                self.__surface_mass_density_water / fvf_water_sample
                for fvf_water_sample in fvf_water
            ]
        )

    def get_keys(self) -> np.ndarray:
        """Returns all primary keys.

        Since this is water, there is no dependency on any ratio.
        Hence, this method returns a list holding a single float of value 0.0.

        """
        return np.zeros(1)

    def get_independents(self) -> np.ndarray:
        """Returns all independent pressure values (Pw).

        Since this is water, this does return with only one single pressure value,
        the reference pressure.
        """
        return np.array(
            [
                self.__pw_ref,
            ]
        )


class Water(FluidImplementation):
    """Class for storing PVT tables for water

    Holds a list of regions (one per PVT table).

    Attributes:
        surface_mass_densities: Surface mass densities
        keep_unit_system: True if the original unit system was kept
        original_unit_system: An ErtEclUnitEnum representing the original unit system
    """

    def __init__(
        self,
        raw: EclPropertyTableRawData,
        unit_system: int,
        surface_mass_densities: np.ndarray,
        keep_unit_system: bool = True,
    ):
        """Initializes a Water object.

        Args:
            raw: Eclipse raw data object
            unit_system: The original unit system
            surface_mass_densities: Surface mass densities
            keep_unit_system:
                True if the original unit system shall be kept,
                False if units shall be converted to SI units.

        """
        super().__init__(keep_unit_system)
        self.surface_mass_densities = surface_mass_densities
        self.original_unit_system = EclUnitEnum(unit_system)
        self.create_water(raw, unit_system)

    def formation_volume_factor_unit(self, latex: bool = False) -> str:
        """Creates and returns the unit symbol of the formation volume factor.

        Args:
            latex: True if the unit symbol shall be returned as LaTeX, False if not.

        Returns:
            The unit symbol of the formation volume factor.

        """
        unit_system = EclUnits.create_unit_system(
            self.original_unit_system
            if self.keep_unit_system
            else EclUnitEnum.ECL_SI_UNITS
        )

        if latex:
            return (
                rf"${{{unit_system.reservoir_volume().symbol}}}"
                rf"/{{{unit_system.surface_volume_liquid().symbol}}}$"
            )
        return (
            f"{unit_system.reservoir_volume().symbol}"
            f"/{unit_system.surface_volume_liquid().symbol}"
        )

    def ratio_unit(self, latex: bool = False) -> str:
        """Creates and returns the unit symbol of the phase ratio.

        Args:
            latex: True if the unit symbol shall be returned as LaTeX, False if not.

        Returns:
            The unit symbol of the phase ratio.

        """
        raise NotImplementedError("Water does not have a phase ratio unit.")

    def water_unit_converter(
        self, unit_system: Union[int, EclUnits.UnitSystem]
    ) -> ConvertUnits:
        """Creates a ConvertUnits object for unit conversions for water.

        Args:
            unit_system:
                Either an integer or an enum
                describing the unit system the units are stored in

        Returns:
            ConvertUnits object for unit conversions.

        """
        if not isinstance(unit_system, EclUnits.UnitSystem):
            unit_system = EclUnits.create_unit_system(unit_system)

        return super().pvdx_unit_converter() or ConvertUnits(
            CreateUnitConverter.ToSI.pressure(unit_system),
            [
                CreateUnitConverter.ToSI.recip_fvf(unit_system),
                CreateUnitConverter.ToSI.compressibility(unit_system),
                CreateUnitConverter.ToSI.recip_fvf_visc(unit_system),
                CreateUnitConverter.ToSI.compressibility(unit_system),
            ],
        )

    def create_water(self, raw: EclPropertyTableRawData, unit_system: int) -> None:
        """Creates interpolants for water from the given raw Eclipse data and uses
        a water unit converter based on the given unit system.

        Args:
            raw: Eclipse raw data object
            unit_system: Integer representation of the unit system used in Eclipse data

        """
        # Holding raw.num_tables values
        ret: List[PvxOBase] = []

        cvrt = self.water_unit_converter(unit_system)

        ret = self.make_interpolants_from_raw_data(
            raw,
            lambda table_index, raw: WaterImpl(
                table_index, raw, self.surface_mass_densities[table_index], cvrt
            ),
        )

        if len(self.surface_mass_densities) != len(ret):
            raise ValueError(
                (
                    "The given Eclipse INIT file seems to be broken"
                    "(number of surface density values does not equal number of PVT regions)."
                )
            )

        self._regions = ret

    @staticmethod
    def from_ecl_init_file(
        ecl_init_file: EclFile, keep_unit_system: bool = False
    ) -> Optional["Water"]:
        """Reads the given Eclipse file and creates a Water object from its data.

        Args:
            ecl_init_file: Eclipse INIT file
            keep_unit_system:
                Set to True if the unit system used in the Eclipse file
                shall be kept, False if SI shall be used.

        Returns:
            A Water object or None if the data in the Eclipse file was invalid

        """
        intehead = ecl_init_file.__getitem__(InitFileDefinitions.INTEHEAD_KW)
        intehead_phase = intehead[InitFileDefinitions.INTEHEAD_PHASE_INDEX]

        if (intehead_phase & (1 << 2)) == 0:
            return None

        raw = EclPropertyTableRawData()

        tab_dims = ecl_init_file.__getitem__("TABDIMS")
        tab = ecl_init_file.__getitem__("TAB")

        raw.num_primary = 1  # Single record per region
        raw.num_rows = 1  # Single record per region
        raw.num_cols = 5  # [ Pw, 1/B, Cw, 1/(B*mu), Cw - Cv ]
        raw.num_tables = tab_dims[InitFileDefinitions.TABDIMS_NTPVTW_ITEM]

        if raw.num_tables == 0:
            return None

        num_tab_elements = (
            raw.num_primary * raw.num_rows * raw.num_cols * raw.num_tables
        )
        start = tab_dims[InitFileDefinitions.TABDIMS_IBPVTW_OFFSET_ITEM] - 1
        raw.data = tab[start : start + num_tab_elements]

        surface_mass_densities = surface_mass_density(
            ecl_init_file, EclPhaseIndex.AQUA, keep_unit_system
        )

        return Water(
            raw,
            intehead[InitFileDefinitions.INTEHEAD_UNIT_INDEX],
            surface_mass_densities,
            keep_unit_system,
        )



from typing import Any, Dict, List, Optional, Union

import pandas as pd
import plotly.graph_objects as go

from webviz_subsurface._abbreviations.number_formatting import si_prefixed

from ._tornado_data import TornadoData


class TornadoBarChart:
    """Creates a plotly bar figure from a TornadoData instance"""

    # pylint: disable=too-many-arguments, too-many-instance-attributes
    def __init__(
        self,
        tornado_data: TornadoData,
        plotly_theme: Dict[str, Any],
        figure_height: Optional[float] = None,
        label_options: str = "detailed",
        locked_si_prefix: Optional[Union[int, str]] = None,
        number_format: str = "",
        unit: str = "",
        spaced: bool = True,
        use_true_base: bool = False,
        show_realization_points: bool = True,
        show_reference: bool = True,
        color_by_sensitivity: bool = False,
        sensitivity_color_map: dict = None,
    ) -> None:
        self._tornadotable = tornado_data.tornadotable
        self._realtable = tornado_data.real_df
        self._reference_average = tornado_data.reference_average
        self._plotly_theme = plotly_theme
        self._number_format = number_format
        self._unit = unit
        self._spaced = spaced
        self._locked_si_prefix = locked_si_prefix
        self._locked_si_prefix_relative: Optional[Union[str, int]]
        self._scale = tornado_data.scale
        self._use_true_base = use_true_base
        self._show_reference = show_reference
        if self._scale == "Percentage":
            self._unit_x = "%"
            self._locked_si_prefix_relative = 0
        else:
            self._unit_x = self._unit
            self._locked_si_prefix_relative = locked_si_prefix
        self._figure_height = figure_height
        self._label_options = label_options
        self._show_scatter = show_realization_points
        self._color_by_sens = color_by_sensitivity
        self._sens_color_map = sensitivity_color_map

    def create_color_list(self, sensitivities: list) -> list:
        return (
            [self._sens_color_map.get(sensname, "grey") for sensname in sensitivities]
            if self._sens_color_map is not None
            else self._plotly_theme["layout"]["colorway"]
        )

    @property
    def figure_height(self) -> Optional[float]:
        """Set height of figure as a function of number of senscases(bars)"""
        return self._figure_height

    def _set_si_prefix(self, value: float) -> str:
        return str(
            si_prefixed(
                value,
                self._number_format,
                self._unit,
                self._spaced,
                self._locked_si_prefix,
            )
        )

    def _set_si_prefix_relative(self, value: float) -> str:
        return str(
            si_prefixed(
                value,
                self._number_format,
                self._unit_x,
                self._spaced,
                self._locked_si_prefix_relative,
            )
        )

    def bar_labels(self, case: str) -> List:
        if self._label_options not in ["simple", "detailed"]:
            return []

        barlabels = []
        for _, row in self._tornadotable.iterrows():
            # combine label if both bars on same side of reference
            comb_label = any((row["low_base"] > 0, row["high_base"] < 0))
            if comb_label:
                xvals = "  |  ".join(
                    [
                        self._set_si_prefix_relative(row["low_tooltip"]),
                        self._set_si_prefix_relative(row["high_tooltip"]),
                    ]
                )
                truevals = "  |  ".join(
                    [
                        self._set_si_prefix(row["true_low"]),
                        self._set_si_prefix(row["true_high"]),
                    ]
                )
                casename = f"{row['low_label']}  |  {row['high_label']}"
            else:
                xvals = self._set_si_prefix_relative(row[f"{case}_tooltip"])
                truevals = self._set_si_prefix(row[f"true_{case}"])
                casename = row[f"{case}_label"]

            text = f"<b>{xvals}</b>, " + ("<br>" if comb_label else "")
            if self._label_options == "detailed":
                text += f"True: {truevals}, <br><b>Case: {casename}</b> "
            barlabels.append(text)
        return barlabels

    def hover_labels(self) -> List:
        hovertext = []
        for _, row in self._tornadotable.iterrows():
            text = f"<b>Sensname: {row['sensname']}</b><br>"
            if row["low_label"] is not None:
                val = row["true_low"] if self._use_true_base else row["low_tooltip"]
                text += (
                    f"{row['low_label']}: <b>{self._set_si_prefix_relative(val)}</b> "
                )
            if row["high_label"] is not None:
                val = row["true_high"] if self._use_true_base else row["high_tooltip"]
                text += (
                    f"{row['high_label']}: <b>{self._set_si_prefix_relative(val)}</b>"
                )
            hovertext.append(text)
        return hovertext

    @property
    def data(self) -> List:
        colors = self.create_color_list(self._tornadotable["sensname"].unique())
        return [
            dict(
                type="bar",
                y=self._tornadotable["sensname"],
                x=self._tornadotable["low"],
                name="low",
                base=self._tornadotable["low_base"]
                if not self._use_true_base
                else (self._reference_average + self._tornadotable["low_base"]),
                customdata=self._tornadotable["low_reals"],
                text=self.bar_labels("low"),
                textposition="auto",
                insidetextanchor="middle",
                hoverinfo="text",
                hovertext=self.hover_labels(),
                orientation="h",
                marker={
                    "line": {"width": 1.5, "color": "black"},
                    "color": colors if self._color_by_sens else None,
                },
            ),
            dict(
                type="bar",
                y=self._tornadotable["sensname"],
                x=self._tornadotable["high"],
                name="high",
                base=self._tornadotable["high_base"]
                if not self._use_true_base
                else (self._reference_average + self._tornadotable["high_base"]),
                customdata=self._tornadotable["high_reals"],
                text=self.bar_labels("high"),
                textposition="auto",
                insidetextanchor="middle",
                hoverinfo="text",
                hovertext=self.hover_labels(),
                orientation="h",
                marker={
                    "line": {"width": 1.5, "color": "black"},
                    "color": colors if self._color_by_sens else None,
                },
            ),
        ]

    def calculate_scatter_value(self, case_values: pd.Series) -> List:
        if self._use_true_base:
            return case_values
        if self._scale == "Percentage":
            return (
                (case_values - self._reference_average) / self._reference_average
            ) * 100
        return case_values - self._reference_average

    @property
    def scatter_data(self) -> List[Dict]:
        return [
            {
                "type": "scatter",
                "mode": "markers",
                "y": df["sensname"],
                "x": self.calculate_scatter_value(df["VALUE"]),
                "text": df["REAL"],
                "hovertemplate": "REAL: <b>%{text}</b><br>"
                + "X: <b>%{x:.1f}</b> <extra></extra>",
                "marker": {
                    "size": 15,
                    "color": self._plotly_theme["layout"]["colorway"][0]
                    if case == "low"
                    else self._plotly_theme["layout"]["colorway"][1],
                },
            }
            for case, df in self._realtable.groupby("case")
        ]

    @property
    def range(self) -> List[float]:
        """Calculate x-axis range so that the reference is centered"""
        max_val = max(self._tornadotable[["low_tooltip", "high_tooltip"]].abs().max())
        if self._use_true_base:
            return [
                self._reference_average - (max_val * 1.1),
                self._reference_average + (max_val * 1.1),
            ]
        return [-max_val * 1.1, max_val * 1.1]

    @property
    def layout(self) -> Dict:
        _layout: Dict[str, Any] = go.Layout()
        _layout.update(self._plotly_theme["layout"])
        _layout.update(
            {
                "height": self.figure_height,
                "barmode": "overlay",
                "margin": {"l": 0, "r": 0, "b": 20, "t": 0, "pad": 21},
                "xaxis": {
                    "title": {
                        "text": self._scale,
                        "standoff": 40,
                    },
                    "range": self.range,
                    "autorange": self._show_scatter or self._tornadotable.empty,
                    "gridwidth": 1,
                    "gridcolor": "whitesmoke",
                    "showgrid": True,
                    "zeroline": False,
                    "linecolor": "black",
                    "showline": True,
                    "automargin": True,
                    "side": "top",
                    "tickfont": {"size": 15},
                },
                "yaxis": {
                    "autorange": True,
                    "showgrid": False,
                    "zeroline": False,
                    "showline": False,
                    "automargin": True,
                    "title": None,
                    "dtick": 1,
                    "tickfont": {"size": 15},
                },
                "showlegend": False,
                "hovermode": "closest",
                "hoverlabel": {"bgcolor": "white", "font_size": 16},
                "annotations": [
                    {
                        "x": 0 if not self._use_true_base else self._reference_average,
                        "y": 1.05,
                        "xref": "x",
                        "yref": "paper",
                        "text": f"<b>{self._set_si_prefix(self._reference_average)}</b>"
                        " (Ref avg)",
                        "showarrow": False,
                        "align": "center",
                        "standoff": 16,
                    }
                ]
                if self._show_reference
                else None,
                "shapes": [
                    {
                        "type": "line",
                        "line": {"width": 3, "color": "lightgrey"},
                        "x0": 0 if not self._use_true_base else self._reference_average,
                        "x1": 0 if not self._use_true_base else self._reference_average,
                        "y0": 0,
                        "y1": 1,
                        "xref": "x",
                        "yref": "y domain",
                    }
                ],
            }
        )
        return _layout

    @property
    def figure(self) -> go.Figure:
        fig = go.Figure({"data": self.data, "layout": self.layout})
        if self._show_scatter:
            fig.update_traces(marker_opacity=0.4, text=None)
            for trace in self.scatter_data:
                fig.add_trace(trace)
        return fig



from webviz_subsurface_components import VectorDefinition, VectorDefinitions

from webviz_subsurface._abbreviations.reservoir_simulation import (
    simulation_vector_description,
)


def test_simulation_vector_description_existing_vector() -> None:
    # Verify test vector exist in VectorDefinitions
    assert "WOPT" in VectorDefinitions

    # Test WITHOUT vector definitions argument
    assert simulation_vector_description("WOPT:A1") == "Oil Production Total, well A1"
    assert (
        simulation_vector_description("PER_DAY_WOPT:A1")
        == "Average Oil Production Total Per day, well A1"
    )
    assert (
        simulation_vector_description("PER_INTVL_WOPT:A1")
        == "Interval Oil Production Total, well A1"
    )
    assert (
        simulation_vector_description("AVG_WOPT:A1")
        == "Average Oil Production Total, well A1"
    )
    assert (
        simulation_vector_description("INTVL_WOPT:A1")
        == "Interval Oil Production Total, well A1"
    )

    # Test WITH vector definitions argument
    vector_definitions = {
        "WOPT": VectorDefinition(type="well", description="Test Description"),
    }
    assert (
        simulation_vector_description("WOPT:A1", vector_definitions)
        == "Test Description, well A1"
    )
    assert (
        simulation_vector_description("PER_DAY_WOPT:A1", vector_definitions)
        == "Average Test Description Per day, well A1"
    )
    assert (
        simulation_vector_description("PER_INTVL_WOPT:A1", vector_definitions)
        == "Interval Test Description, well A1"
    )
    assert (
        simulation_vector_description("AVG_WOPT:A1", vector_definitions)
        == "Average Test Description, well A1"
    )
    assert (
        simulation_vector_description("INTVL_WOPT:A1", vector_definitions)
        == "Interval Test Description, well A1"
    )


def test_simulation_vector_description_non_existing_vector() -> None:
    # Verify test vector does not exist in VectorDefinitions
    assert "Custom" not in VectorDefinitions

    # Test WITHOUT vector definitions argument
    assert simulation_vector_description("Custom:A1") == "Custom:A1"
    assert (
        simulation_vector_description("PER_DAY_Custom:A1")
        == "Average Custom:A1 Per day"
    )
    assert simulation_vector_description("PER_INTVL_Custom:A1") == "Interval Custom:A1"
    assert simulation_vector_description("AVG_Custom:A1") == "Average Custom:A1"
    assert simulation_vector_description("INTVL_Custom:A1") == "Interval Custom:A1"

    # Test WITH vector definitions argument
    vector_definitions = {
        "Custom": VectorDefinition(type="field", description="Custom Description"),
    }
    assert (
        simulation_vector_description("Custom:A1", vector_definitions)
        == "Custom Description, field A1"
    )
    assert (
        simulation_vector_description("PER_DAY_Custom:A1", vector_definitions)
        == "Average Custom Description Per day, field A1"
    )
    assert (
        simulation_vector_description("PER_INTVL_Custom:A1", vector_definitions)
        == "Interval Custom Description, field A1"
    )
    assert (
        simulation_vector_description("AVG_Custom:A1", vector_definitions)
        == "Average Custom Description, field A1"
    )
    assert (
        simulation_vector_description("INTVL_Custom:A1", vector_definitions)
        == "Interval Custom Description, field A1"
    )



from dataclasses import dataclass
from typing import Dict

import numpy as np
import pyarrow as pa

from ._field_metadata import is_rate_from_field_meta
from .ensemble_summary_provider import Frequency


def _truncate_day_to_monday(datetime_day: np.datetime64) -> np.datetime64:
    # A bit hackish, utilizes the fact that datetime64 is relative to epoch
    # 1970-01-01 which is a Thursday
    return datetime_day.astype("datetime64[W]").astype("datetime64[D]") + 4


def _quarter_start_month(datetime_day: np.datetime64) -> np.datetime64:
    # A bit hackish, utilizes the fact that datetime64 is relative to epoch
    # 1970-01-01 which is the first day in Q1.
    datetime_month = np.datetime64(datetime_day, "M")
    return datetime_month - (datetime_month.astype(int) % 3)


def generate_normalized_sample_dates(
    min_date: np.datetime64, max_date: np.datetime64, freq: Frequency
) -> np.ndarray:
    """Returns array of normalized sample dates to cover the min_date to max_date
    range with the specified frequency.
    The return numpy array will have sample dates with dtype datetime64[ms]
    """

    if freq == Frequency.DAILY:
        start = np.datetime64(min_date, "D")
        stop = np.datetime64(max_date, "D")
        if stop < max_date:
            stop += 1
        sampledates = np.arange(start, stop + 1)
    elif freq == Frequency.WEEKLY:
        start = _truncate_day_to_monday(np.datetime64(min_date, "D"))
        stop = _truncate_day_to_monday(np.datetime64(max_date, "D"))
        if start > min_date:
            start -= 7
        if stop < max_date:
            stop += 7
        sampledates = np.arange(start, stop + 1, 7)
    elif freq == Frequency.MONTHLY:
        start = np.datetime64(min_date, "M")
        stop = np.datetime64(max_date, "M")
        if stop < max_date:
            stop += 1
        sampledates = np.arange(start, stop + 1)
    elif freq == Frequency.QUARTERLY:
        start = _quarter_start_month(min_date)
        stop = _quarter_start_month(max_date)
        if stop < max_date:
            stop += 3
        sampledates = np.arange(start, stop + 1, 3)
    elif freq == Frequency.YEARLY:
        start = np.datetime64(min_date, "Y")
        stop = np.datetime64(max_date, "Y")
        if stop < max_date:
            stop += 1
        sampledates = np.arange(start, stop + 1)
    else:
        raise NotImplementedError(
            f"Currently not supporting resampling to frequency {freq}."
        )

    sampledates = sampledates.astype("datetime64[ms]")

    return sampledates


def interpolate_backfill(
    x: np.ndarray, xp: np.ndarray, yp: np.ndarray, yleft: float, yright: float
) -> np.ndarray:
    # pylint: disable=invalid-name
    """Do back-filling interpolation of the coordinates in xp and yp, evaluated at the
    x-coordinates specified in x.
    Note that xp and yp must be arrays of the same length.
    It is assumed that both the x and the xp array is sorted in increasing order.
    """

    # Finds the leftmost valid insertion indices for the values x in xp
    indices = np.searchsorted(xp, x, side="left")

    padded_y = np.concatenate((yp, np.array([yright])))

    ret_arr = padded_y[indices]

    if x[0] < xp[0]:
        idx = np.searchsorted(x, xp[0])
        ret_arr[0:idx] = yleft

    return ret_arr


def resample_single_real_table(table: pa.Table, freq: Frequency) -> pa.Table:
    """Resample table that contains only a single realization.
    The table must contain a DATE column and it must be sorted on DATE
    """

    # Notes:
    # Getting meta data using json.loads() takes quite a bit of time!!
    # We should provide this info in another way.

    schema = table.schema

    raw_dates_np = table.column("DATE").to_numpy()
    raw_dates_np_as_uint = raw_dates_np.astype(np.uint64)

    min_raw_date = np.min(raw_dates_np)
    max_raw_date = np.max(raw_dates_np)

    sample_dates_np = generate_normalized_sample_dates(
        min_raw_date, max_raw_date, freq=freq
    )
    sample_dates_np_as_uint = sample_dates_np.astype(np.uint64)

    column_arrays = []

    for colname in schema.names:
        if colname == "DATE":
            column_arrays.append(sample_dates_np)
        elif colname == "REAL":
            column_arrays.append(
                np.full(len(sample_dates_np), table.column("REAL")[0].as_py())
            )
        else:
            raw_numpy_arr = table.column(colname).to_numpy()
            if is_rate_from_field_meta(table.field(colname)):
                i = interpolate_backfill(
                    sample_dates_np_as_uint, raw_dates_np_as_uint, raw_numpy_arr, 0, 0
                )
            else:
                i = np.interp(
                    sample_dates_np_as_uint, raw_dates_np_as_uint, raw_numpy_arr
                )

            column_arrays.append(i)

    ret_table = pa.table(column_arrays, schema=schema)

    return ret_table


@dataclass
class RealInterpolationInfo:
    raw_dates_np: np.ndarray
    raw_dates_np_as_uint: np.ndarray
    sample_dates_np: np.ndarray
    sample_dates_np_as_uint: np.ndarray


def _extract_real_interpolation_info(
    table: pa.Table, start_row_idx: int, row_count: int, freq: Frequency
) -> RealInterpolationInfo:

    real_dates = table["DATE"].slice(start_row_idx, row_count).to_numpy()

    min_raw_date = np.min(real_dates)
    max_raw_date = np.max(real_dates)
    sample_dates = generate_normalized_sample_dates(min_raw_date, max_raw_date, freq)

    return RealInterpolationInfo(
        raw_dates_np=real_dates,
        raw_dates_np_as_uint=real_dates.astype(np.uint64),
        sample_dates_np=sample_dates,
        sample_dates_np_as_uint=sample_dates.astype(np.uint64),
    )


def resample_segmented_multi_real_table(table: pa.Table, freq: Frequency) -> pa.Table:
    """Resample table containing multiple realizations.
    The table must contain both a REAL and a DATE column.
    The table must be segmented on REAL (so that all rows from a single
    realization are contiguous) and within each REAL segment, it must be
    sorted on DATE.
    The segmentation is needed since interpolations must be done per realization
    and we utilize slicing on rows for speed.
    """
    # pylint: disable=too-many-locals

    real_arr_np = table.column("REAL").to_numpy()
    unique_reals, first_occurrence_idx, real_counts = np.unique(
        real_arr_np, return_index=True, return_counts=True
    )

    output_columns_dict: Dict[str, pa.ChunkedArray] = {}

    real_interpolation_info_dict: Dict[int, RealInterpolationInfo] = {}

    for colname in table.schema.names:
        if colname in ["DATE", "REAL"]:
            continue

        is_rate = is_rate_from_field_meta(table.field(colname))
        raw_whole_numpy_arr = table.column(colname).to_numpy()

        vec_arr_list = []
        for i, real in enumerate(unique_reals):
            start_row_idx = first_occurrence_idx[i]
            row_count = real_counts[i]

            rii = real_interpolation_info_dict.get(real)
            if not rii:
                rii = _extract_real_interpolation_info(
                    table, start_row_idx, row_count, freq
                )
                real_interpolation_info_dict[real] = rii

            raw_numpy_arr = raw_whole_numpy_arr[
                start_row_idx : start_row_idx + row_count
            ]

            if is_rate:
                inter = interpolate_backfill(
                    rii.sample_dates_np_as_uint,
                    rii.raw_dates_np_as_uint,
                    raw_numpy_arr,
                    0,
                    0,
                )
            else:
                inter = np.interp(
                    rii.sample_dates_np_as_uint,
                    rii.raw_dates_np_as_uint,
                    raw_numpy_arr,
                )

            arr_length = len(rii.sample_dates_np_as_uint)
            assert arr_length == len(inter)

            vec_arr_list.append(inter)

        output_columns_dict[colname] = pa.chunked_array(vec_arr_list)

    date_arr_list = []
    real_arr_list = []
    for real in unique_reals:
        rii = real_interpolation_info_dict[real]
        arr_length = len(rii.sample_dates_np)
        date_arr_list.append(rii.sample_dates_np)
        real_arr_list.append(np.full(arr_length, real))

    output_columns_dict["DATE"] = pa.chunked_array(date_arr_list)
    output_columns_dict["REAL"] = pa.chunked_array(real_arr_list)

    ret_table = pa.table(output_columns_dict, schema=table.schema)

    return ret_table


def _compute_interpolation_weight(
    d: np.datetime64, d0: np.datetime64, d1: np.datetime64
) -> float:
    # pylint: disable=invalid-name
    d_as_uint = d.astype(np.uint64)
    d0_as_uint = d0.astype(np.uint64)
    d1_as_uint = d1.astype(np.uint64)
    return float(d_as_uint - d0_as_uint) / float(d1_as_uint - d0_as_uint)


def sample_segmented_multi_real_table_at_date(
    table: pa.Table, np_datetime: np.datetime64
) -> pa.Table:
    """Sample table containing multiple realizations at the specified date.
    The table must contain both a REAL and a DATE column.
    The table must be segmented on REAL (so that all rows from a single
    realization are contiguous) and within each REAL segment, it must be
    sorted on DATE.
    """
    # pylint: disable=too-many-locals

    unique_reals_arr_np, first_occurrence_idx, real_counts = np.unique(
        table.column("REAL").to_numpy(), return_index=True, return_counts=True
    )

    all_dates_arr_np = table.column("DATE").to_numpy()

    # Will receive row indices into the full input table for the two values we should
    # interpolate/blend between.
    # To keep things simple we always add two indices for each realization even if
    # we know that no interpolation will be needed (e.g. exact matches)
    row_indices = []

    # Will receive the blending weights for doing interpolation
    interpolate_t_arr = np.zeros(len(unique_reals_arr_np))

    # Array with mask for selecting values when doing backfill. A value of 1 will select
    # v1, while a value of 0 will yield a 0 value
    backfill_mask_arr = np.ones(len(unique_reals_arr_np))

    for i, _real in enumerate(unique_reals_arr_np):
        # Starting row of this realization and number of rows belonging to realization
        start_row_idx = first_occurrence_idx[i]
        row_count = real_counts[i]

        # Get slice of the dates for just this realization
        dates_arr_np = all_dates_arr_np[start_row_idx : start_row_idx + row_count]
        assert len(dates_arr_np) > 0

        # OUTSIDE RANGE (query date is before our first date)
        if np_datetime < dates_arr_np[0]:
            row_indices.append(start_row_idx)
            row_indices.append(start_row_idx)
            # Extrapolate or just fill with 0 for rates
            # interpolate_t_arr[i] = 0
            backfill_mask_arr[i] = 0

        # OUTSIDE RANGE (query date is beyond our last date)
        elif np_datetime > dates_arr_np[-1]:
            row_indices.append(start_row_idx + row_count - 1)
            row_indices.append(start_row_idx + row_count - 1)
            # Extrapolate or just fill with 0 for rates. For interpolation, t should
            # really 1, but since the rows are duplicated it does not matter
            # interpolate_t_arr[i] = 0
            backfill_mask_arr[i] = 0

        # EXACT MATCH on the LAST DATE
        elif np_datetime == dates_arr_np[-1]:
            row_indices.append(start_row_idx + row_count - 1)
            row_indices.append(start_row_idx + row_count - 1)
            # interpolate_t_arr[i] = 0
            # backfill_mask_arr[i] = 1

        else:
            # Search for query date amongst the realization's dates.
            # last_insertion_index is the last legal insertion index of the queried value
            last_insertion_index: int = np.searchsorted(
                dates_arr_np, np_datetime, side="right"
            ).item()

            assert 0 < last_insertion_index < len(dates_arr_np)
            assert dates_arr_np[last_insertion_index - 1] <= np_datetime
            assert dates_arr_np[last_insertion_index] > np_datetime

            if dates_arr_np[last_insertion_index - 1] == np_datetime:
                # Exact match
                row_indices.append(start_row_idx + last_insertion_index - 1)
                row_indices.append(start_row_idx + last_insertion_index - 1)
                # interpolate_t_arr[i] = 0
                # backfill_mask_arr[i] = 1
            else:
                row_indices.append(start_row_idx + last_insertion_index - 1)
                row_indices.append(start_row_idx + last_insertion_index)
                interpolate_t_arr[i] = _compute_interpolation_weight(
                    np_datetime,
                    dates_arr_np[last_insertion_index - 1],
                    dates_arr_np[last_insertion_index],
                )
                # backfill_mask_arr[i] = 1

    column_arrays = []

    for colname in table.schema.names:
        if colname == "REAL":
            column_arrays.append(unique_reals_arr_np)
        elif colname == "DATE":
            column_arrays.append(np.full(len(unique_reals_arr_np), np_datetime))
        else:
            records_np = table.column(colname).take(row_indices).to_numpy()
            if is_rate_from_field_meta(table.field(colname)):
                v1_arr = records_np[1::2]
                interpolated_vec_values = v1_arr * backfill_mask_arr
            else:
                v0_arr = records_np[0::2]
                v1_arr = records_np[1::2]
                delta_arr = v1_arr - v0_arr
                interpolated_vec_values = v0_arr + (delta_arr * interpolate_t_arr)

            column_arrays.append(pa.array(interpolated_vec_values))

    ret_table = pa.table(column_arrays, schema=table.schema)

    return ret_table



from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple, Union

import pandas as pd
from dash import Dash, dcc
from webviz_config import WebvizConfigTheme, WebvizPluginABC, WebvizSettings
from webviz_config.deprecation_decorators import deprecated_plugin_arguments
from webviz_config.webviz_instance_info import WEBVIZ_INSTANCE_INFO, WebvizRunMode

from webviz_subsurface._models import (
    EnsembleSetModel,
    caching_ensemble_set_model_factory,
)
from webviz_subsurface._providers import (
    EnsembleSummaryProviderFactory,
    EnsembleTableProvider,
    EnsembleTableProviderFactory,
    Frequency,
)

from .controllers.property_delta_controller import property_delta_controller
from .controllers.property_qc_controller import property_qc_controller
from .controllers.property_response_controller import property_response_controller
from .data_loaders import read_csv
from .models import (
    PropertyStatisticsModel,
    ProviderTimeSeriesDataModel,
    SimulationTimeSeriesModel,
)
from .utils.surface import generate_surface_table, get_path
from .views.main_view import main_view


def check_deprecation_argument(
    csvfile_statistics: Optional[Path], csvfile_smry: Optional[Path]
) -> Optional[Tuple[str, str]]:
    if any(elm is not None for elm in [csvfile_statistics, csvfile_smry]):
        return (
            "The usage of aggregated csvfiles as user input options are deprecated. "
            "Please provide feedback if you see a need for a continuation "
            "of this functionality ",
            "",
        )
    return None


class PropertyStatistics(WebvizPluginABC):
    """This plugin visualizes ensemble statistics calculated from grid properties.

---
**The main input to this plugin is property statistics extracted from grid models.
See the documentation in [fmu-tools](http://fmu-docs.equinor.com/) on how to generate this data.
Additional data includes UNSMRY data and optionally irap binary surfaces stored in standardized \
FMU format.


**Using raw ensemble data stored in realization folders**
* **`ensembles`:** Which ensembles in `shared_settings` to visualize.
* **`rel_file_pattern`:** path to `.arrow` files with summary data.
* **`statistic_file`:** Csv file for each realization with property statistics. See the \
    documentation in [fmu-tools](http://fmu-docs.equinor.com/) on how to generate this data.
* **`column_keys`:** List of vectors to extract. If not given, all vectors \
    from the simulations will be extracted. Wild card asterisk `*` can be used.
* **`time_index`:** Time separation between extracted values. Can be e.g. `monthly` (default) or \
    `yearly`.
* **`surface_renaming`:** Optional dictionary to rename properties/zones to match filenames \
    stored on FMU standardized format (zone--property.gri)

---

?> `Arrow` format for simulation time series data can be generated using the `ECL2CSV` forward \
model in ERT. On existing ensembles the command line tool `smry2arrow_batch` can be used to \
generate arrow files.

?> Folders with statistical surfaces are assumed located at \
`<ensemble_path>/share/results/maps/<ensemble>/<statistic>` where `statistic` are subfolders \
with statistical calculation: `mean`, `stddev`, `p10`, `p90`, `min`, `max`.

!> For smry data it is **strongly recommended** to keep the data frequency to a regular frequency \
(like `monthly` or `yearly`). This applies to both csv input and when reading from `UNSMRY` \
(controlled by the `sampling` key). This is because the statistics and fancharts are calculated \
per DATE over all realizations in an ensemble, and the available dates should therefore not \
differ between individual realizations of an ensemble.


"""

    # pylint: disable=too-many-arguments, too-many-locals
    @deprecated_plugin_arguments(check_deprecation_argument)
    def __init__(
        self,
        app: Dash,
        webviz_settings: WebvizSettings,
        ensembles: Optional[list] = None,
        rel_file_pattern: str = "share/results/unsmry/*.arrow",
        statistics_file: str = "share/results/tables/gridpropstatistics.csv",
        surface_renaming: Optional[dict] = None,
        time_index: str = "monthly",
        column_keys: Optional[list] = None,
        csvfile_statistics: Path = None,
        csvfile_smry: Path = None,
    ):
        super().__init__()
        self.theme: WebvizConfigTheme = webviz_settings.theme
        self.ensembles = ensembles
        self._surface_folders: Union[dict, None] = None
        self._vmodel: Optional[
            Union[SimulationTimeSeriesModel, ProviderTimeSeriesDataModel]
        ] = None
        run_mode_portable = WEBVIZ_INSTANCE_INFO.run_mode == WebvizRunMode.PORTABLE
        table_provider_factory = EnsembleTableProviderFactory.instance()

        if ensembles is not None:
            ensemble_paths = {
                ensemble_name: webviz_settings.shared_settings["scratch_ensembles"][
                    ensemble_name
                ]
                for ensemble_name in ensembles
            }

            resampling_frequency = Frequency(time_index)
            provider_factory = EnsembleSummaryProviderFactory.instance()

            try:
                provider_set = {
                    ens: provider_factory.create_from_arrow_unsmry_presampled(
                        str(ens_path), rel_file_pattern, resampling_frequency
                    )
                    for ens, ens_path in ensemble_paths.items()
                }
                self._vmodel = ProviderTimeSeriesDataModel(
                    provider_set=provider_set, column_keys=column_keys
                )
                table_provider_set = {
                    ens: table_provider_factory.create_from_per_realization_csv_file(
                        ens_path, statistics_file
                    )
                    for ens, ens_path in ensemble_paths.items()
                }
                property_df = create_df_from_table_provider(table_provider_set)

            except ValueError as error:
                message = (
                    f"Some/all ensembles are missing arrow files at {rel_file_pattern}.\n"
                    "If no arrow files have been generated with `ERT` using `ECL2CSV`, "
                    "the commandline tool `smry2arrow_batch` can be used to generate arrow "
                    "files for an ensemble"
                )
                if not run_mode_portable:
                    raise ValueError(message) from error

                # NOTE: this part below is to ensure backwards compatibility for portable app's
                # created before the arrow support. It should be removed in the future.
                emodel: EnsembleSetModel = (
                    caching_ensemble_set_model_factory.get_or_create_model(
                        ensemble_paths=ensemble_paths,
                        time_index=time_index,
                        column_keys=column_keys,
                    )
                )
                self._vmodel = SimulationTimeSeriesModel(
                    dataframe=emodel.get_or_load_smry_cached()
                )
                property_df = emodel.load_csv(csv_file=Path(statistics_file))

            self._surface_folders = {
                ens: Path(ens_path.split("realization")[0]) / "share/results/maps" / ens
                for ens, ens_path in ensemble_paths.items()
            }

        else:
            if csvfile_statistics is None:
                raise ValueError(
                    "If not 'ensembles', then csvfile_statistics must be provided"
                )
            # NOTE: the try/except is for backwards compatibility with existing portable app's.
            # It should be removed in the future together with the support of aggregated csv-files
            try:
                property_df = create_df_from_table_provider(
                    table_provider_factory.create_provider_set_from_aggregated_csv_file(
                        csvfile_statistics
                    )
                )
            except FileNotFoundError:
                if not run_mode_portable:
                    raise
                property_df = read_csv(csvfile_statistics)

            if csvfile_smry is not None:
                try:
                    smry_df = create_df_from_table_provider(
                        table_provider_factory.create_provider_set_from_aggregated_csv_file(
                            csvfile_smry
                        )
                    )
                except FileNotFoundError:
                    if not run_mode_portable:
                        raise
                    smry_df = read_csv(csvfile_smry)

                self._vmodel = SimulationTimeSeriesModel(dataframe=smry_df)

        self._pmodel = PropertyStatisticsModel(dataframe=property_df, theme=self.theme)

        self._surface_renaming = surface_renaming if surface_renaming else {}
        self._surface_table = generate_surface_table(
            zones=self._pmodel.zones,
            properties=self._pmodel.properties,
            ensembles=self._pmodel.ensembles,
            surface_folders=self._surface_folders,
            surface_renaming=self._surface_renaming,
        )
        self.set_callbacks(app)

    @property
    def layout(self) -> dcc.Tabs:
        return main_view(
            get_uuid=self.uuid,
            property_model=self._pmodel,
            surface_folders=self._surface_folders,
            vector_model=self._vmodel,
        )

    def set_callbacks(self, app: Dash) -> None:
        property_qc_controller(app=app, get_uuid=self.uuid, property_model=self._pmodel)
        if len(self._pmodel.ensembles) > 1:
            property_delta_controller(
                app=app,
                get_uuid=self.uuid,
                property_model=self._pmodel,
                surface_table=self._surface_table,
            )
        if self._vmodel is not None:
            property_response_controller(
                app=app,
                get_uuid=self.uuid,
                surface_table=self._surface_table,
                property_model=self._pmodel,
                timeseries_model=self._vmodel,
            )

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        store: List[Tuple[Callable, list]] = [
            (
                generate_surface_table,
                [
                    {
                        "zones": self._pmodel.zones,
                        "properties": self._pmodel.properties,
                        "ensembles": self._pmodel.ensembles,
                        "surface_folders": self._surface_folders,
                        "surface_renaming": self._surface_renaming,
                    }
                ],
            )
        ]
        if self._surface_folders is not None:
            for path in self._surface_table["path"].unique():
                store.append((get_path, [{"path": Path(path)}]))
        return store


def create_df_from_table_provider(
    provider_set: Dict[str, EnsembleTableProvider]
) -> pd.DataFrame:
    """Aggregates parameters from all ensemble into a common dataframe."""
    dfs = []
    for ens_name, provider in provider_set.items():
        df = provider.get_column_data(column_names=provider.column_names())
        df["ENSEMBLE"] = ens_name
        dfs.append(df)
    return pd.concat(dfs)



import fnmatch
import re
from typing import List

from .ensemble_summary_provider import EnsembleSummaryProvider


def get_matching_vector_names(
    provider: EnsembleSummaryProvider, column_keys: List[str]
) -> List[str]:
    """Returns a list of vectors that match the input columns_keys that
    can have unix shell wildcards.

    Example of use:
    column_keys = ["FOPT", "WGOR*"]
    matching_vector_names = get_matching_vector_names(provider, column_keys)
    df = provider.get_vectors_df(matching_vector_names, None)

    """
    regex = re.compile(
        "|".join([fnmatch.translate(col) for col in column_keys]),
        flags=re.IGNORECASE,
    )
    return [vec for vec in provider.vector_names() if regex.fullmatch(vec)]



from typing import List, Union

import pandas as pd
from dash import Input, Output, callback
from webviz_config import WebvizSettings
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import ViewABC

from .._plugin_ids import PluginIds
from ..view_elements import Graph
from ._view_functions import calc_statistics, filter_df, label_map


class LineView(ViewABC):
    class Ids(StrEnum):
        LINE_CHART = "line-chart"
        SETTINGS = "settings"

    def __init__(
        self,
        bhp_df: pd.DataFrame,
        webviz_settings: WebvizSettings,
    ) -> None:
        super().__init__("Line chart")

        self.bhp_df = bhp_df

        column = self.add_column()
        column.add_view_element(Graph(), LineView.Ids.LINE_CHART)
        self.theme = webviz_settings.theme

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.view_element(LineView.Ids.LINE_CHART)
                .component_unique_id(Graph.Ids.GRAPH)
                .to_string(),
                "figure",
            ),
            Input(self.get_store_unique_id(PluginIds.Stores.SELECTED_ENSEMBLE), "data"),
            Input(self.get_store_unique_id(PluginIds.Stores.SELECTED_SORT_BY), "data"),
            Input(
                self.get_store_unique_id(
                    PluginIds.Stores.SELECTED_ASCENDING_DESCENDING
                ),
                "data",
            ),
            Input(
                self.get_store_unique_id(PluginIds.Stores.SELECTED_MAX_NUMBER_OF_WELLS),
                "data",
            ),
            Input(self.get_store_unique_id(PluginIds.Stores.SELECTED_WELLS), "data"),
            Input(
                self.get_store_unique_id(PluginIds.Stores.SELECTED_STATISTICS), "data"
            ),
        )
        def _update_plot(
            ensemble: str,
            sort_by: str,
            ascending: bool,
            n_wells: int,
            wells: Union[str, List[str]],
            stat_bars: Union[str, List[str]],
        ) -> dict:

            wells = wells if isinstance(wells, list) else [wells]
            stat_bars = stat_bars if isinstance(stat_bars, list) else [stat_bars]
            df = filter_df(df=self.bhp_df, ensemble=ensemble, wells=wells)
            stat_df = (
                calc_statistics(df)
                .sort_values(sort_by, ascending=ascending)
                .iloc[0:n_wells, :]
            )
            line_chart = []
            for stat in stat_bars:
                yaxis = "y2" if stat == "count" else "y"
                line_chart.append(
                    {
                        "x": [vec[5:] for vec in stat_df.index],  # strip WBHP:
                        "y": stat_df[stat],
                        "name": [
                            key for key, value in label_map().items() if value == stat
                        ][0],
                        "yaxis": yaxis,
                        "type": "line",
                        "offsetgroup": stat,
                        "showlegend": True,
                    }
                )
            layout = self.theme.create_themed_layout(
                {
                    "yaxis": {
                        "side": "left",
                        "title": "Bottom hole pressure",
                        "showgrid": True,
                    },
                    "yaxis2": {
                        "side": "right",
                        "overlaying": "y",
                        "title": "Count (data points)",
                        "showgrid": False,
                    },
                    "xaxis": {"showgrid": False},
                    "barmode": "group",
                    "legend": {"x": 1.05},
                }
            )
            return {"data": line_chart, "layout": layout}



from typing import List, Optional

import webviz_core_components as wcc
from dash import Input, Output, callback, html
from dash.development.base_component import Component
from dash.exceptions import PreventUpdate
from webviz_config.utils import StrEnum, callback_typecheck
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC

from webviz_subsurface._utils.formatting import printable_int_list

from .._types import StatisticsFromOptions


class FilterRealizationSettings(SettingsGroupABC):
    class Ids(StrEnum):
        REALIZATIONS_FILTER_SPAN = "realization-filter-span"
        STATISTICS_FROM_RADIO_ITEMS = "statistics-from-radio-items"
        REALIZATIONS_FILTER_SELECTOR = "realizations-filter-selector"

    def __init__(self, realizations: List[int]) -> None:
        super().__init__("Filter Realizations")
        self._realizations = realizations

    def layout(self) -> List[Component]:
        return [
            html.Div(
                style={"display": "inline-flex"},
                children=[
                    html.Label(
                        "Realizations: ",
                        style={"font-weight": "bold"},
                    ),
                    html.Label(
                        id=self.register_component_unique_id(
                            FilterRealizationSettings.Ids.REALIZATIONS_FILTER_SPAN
                        ),
                        style={
                            "margin-left": "10px",
                            "margin-bottom": "5px",
                        },
                        children=f"{min(self._realizations)}-{max(self._realizations)}",
                    ),
                ],
            ),
            wcc.RadioItems(
                label="Statistics calculated from:",
                id=self.register_component_unique_id(
                    FilterRealizationSettings.Ids.STATISTICS_FROM_RADIO_ITEMS
                ),
                style={"margin-bottom": "10px"},
                options=[
                    {
                        "label": "All",
                        "value": StatisticsFromOptions.ALL_REALIZATIONS,
                    },
                    {
                        "label": "Selected",
                        "value": StatisticsFromOptions.SELECTED_REALIZATIONS,
                    },
                ],
                value=StatisticsFromOptions.ALL_REALIZATIONS,
                vertical=False,
            ),
            wcc.Select(
                id=self.register_component_unique_id(
                    FilterRealizationSettings.Ids.REALIZATIONS_FILTER_SELECTOR
                ),
                options=[{"label": i, "value": i} for i in self._realizations],
                value=self._realizations,
                size=min(10, len(self._realizations)),
            ),
        ]

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.component_unique_id(
                    FilterRealizationSettings.Ids.REALIZATIONS_FILTER_SPAN
                ).to_string(),
                "children",
            ),
            Input(
                self.component_unique_id(
                    FilterRealizationSettings.Ids.REALIZATIONS_FILTER_SELECTOR
                ).to_string(),
                "value",
            ),
        )
        @callback_typecheck
        def _update_realization_range(realizations: List[int]) -> Optional[str]:
            if not realizations:
                raise PreventUpdate

            realizations_filter_text = printable_int_list(realizations)

            return realizations_filter_text



from typing import List, Union

import webviz_core_components as wcc
from dash import Input, Output, callback
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import ViewABC

from ..._reusable_settings import FilterLayout
from ..._reusable_view_element import GeneralViewElement
from ..._utils import RftPlotterDataModel, filter_frame
from ._settings import Selections
from ._utils import update_misfit_per_real_plot


class MisfitPerRealView(ViewABC):
    class Ids(StrEnum):
        SELECTIONS = "selections"
        FILTERS = "filters"
        VIEW_ELEMENT = "view-element"

    def __init__(self, datamodel: RftPlotterDataModel) -> None:
        super().__init__("Misfit per real")
        self._datamodel = datamodel

        self.add_settings_groups(
            {
                self.Ids.SELECTIONS: Selections(self._datamodel.ensembles),
                self.Ids.FILTERS: FilterLayout(
                    wells=self._datamodel.well_names,
                    zones=self._datamodel.zone_names,
                    dates=self._datamodel.dates,
                ),
            }
        )

        self.add_view_element(GeneralViewElement(), self.Ids.VIEW_ELEMENT)

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.view_element(self.Ids.VIEW_ELEMENT)
                .component_unique_id(GeneralViewElement.Ids.CHART)
                .to_string(),
                "children",
            ),
            Input(
                self.settings_group(self.Ids.SELECTIONS)
                .component_unique_id(Selections.Ids.ENSEMBLES)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(self.Ids.FILTERS)
                .component_unique_id(FilterLayout.Ids.FILTER_WELLS)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(self.Ids.FILTERS)
                .component_unique_id(FilterLayout.Ids.FILTER_ZONES)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(self.Ids.FILTERS)
                .component_unique_id(FilterLayout.Ids.FILTER_DATES)
                .to_string(),
                "value",
            ),
        )
        def _misfit_plot(
            ensembles: List[str], wells: List[str], zones: List[str], dates: List[str]
        ) -> Union[str, List[wcc.Graph]]:
            df = filter_frame(
                self._datamodel.ertdatadf,
                {"WELL": wells, "ZONE": zones, "DATE": dates, "ENSEMBLE": ensembles},
            )
            if df.empty:
                return "No data matching the given filter criterias"

            return update_misfit_per_real_plot(df, self._datamodel.enscolors)



import datetime
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

import numpy as np
import pandas as pd

from webviz_subsurface._models import GruptreeModel, WellAttributesModel
from webviz_subsurface._providers import EnsembleSummaryProvider

from .._types import NodeType, PressurePlotMode


class EnsembleWellAnalysisData:
    """This class holds the summary data provider."""

    def __init__(
        self,
        ensemble_name: str,
        provider: EnsembleSummaryProvider,
        gruptree_model: GruptreeModel,
        well_attributes_model: WellAttributesModel,
        filter_out_startswith: Optional[str] = None,
    ):
        self._ensemble_name = ensemble_name
        self._gruptree_model = gruptree_model
        self._well_attributes_model = well_attributes_model
        self._provider = provider
        self._vector_names = self._provider.vector_names()
        self._realizations = self._provider.realizations()
        self._wells: List[str] = [
            vec.split(":")[1] for vec in self._vector_names if vec.startswith("WOPT:")
        ]
        if filter_out_startswith is not None:
            self._wells = [
                well
                for well in self._wells
                if not well.startswith(filter_out_startswith)
            ]

        well_sumvecs = [vec for vec in self._vector_names if vec.startswith("W")]
        group_sumvecs = [vec for vec in self._vector_names if vec.startswith("GPR:")]
        self._smry = provider.get_vectors_df(well_sumvecs + group_sumvecs, None)

    @property
    def webviz_store(self) -> List[Tuple[Callable, List[Dict]]]:
        return [
            self._gruptree_model.webviz_store,
            self._well_attributes_model.webviz_store,
        ]

    @property
    def summary_data(self) -> pd.DataFrame:
        return self._smry

    @property
    def dates(self) -> List[datetime.datetime]:
        return list(self._smry["DATE"].unique())

    @property
    def realizations(self) -> List[int]:
        return self._realizations

    @property
    def wells(self) -> List[str]:
        return self._wells

    @property
    def well_attributes(self) -> pd.DataFrame:
        return self._well_attributes_model.category_dict

    def filter_on_well_attributes(
        self, well_attributes_filter: Dict[str, List[str]]
    ) -> Set[str]:
        """Filters wells based on the input well attributes filter.

        The input parameter well_attribute_filter has the form:
        {
            "welltype":["oil_producer", "water_injector", ...],
            "category2": ...
        }

        Any categories in the filter than does not exist in the ensemble
        well attributes will be ignored. If no categories match, then all
        wells is returned.
        """
        well_attr_df = self._well_attributes_model.dataframe_melted.fillna("Undefined")
        filtered_wells = set(self._wells)
        for category, values in well_attributes_filter.items():
            if category in self._well_attributes_model.categories:
                df = well_attr_df[
                    (well_attr_df["CATEGORY"] == category)
                    & (well_attr_df["VALUE"].isin(values))
                ]
                filtered_wells = filtered_wells.intersection(set(df["WELL"].unique()))
        return filtered_wells

    def get_summary_data(
        self,
        well_sumvec: str,
        prod_from_date: Union[datetime.datetime, None],
        prod_until_date: Union[datetime.datetime, None],
    ) -> pd.DataFrame:
        """Returns all summary data matching the well_sumvec. If the prod_from_date
        is not None it will return all dates after that date and subtract the cumulative
        production at that date. If prod_until_dates is not None it will filter out all
        dates after that date.
        """
        sumvecs = [f"{well_sumvec}:{well}" for well in self._wells]
        df = self._smry[["REAL", "DATE"] + sumvecs]
        max_date = df["DATE"].max()
        min_date = df["DATE"].min()

        if prod_from_date is not None:
            df = df[df["DATE"] >= prod_from_date]

            # If the prod_from_date exists in the ensemble, subtract the
            # production at that date from all dates.
            if min_date <= prod_from_date <= max_date:
                df_date = df[df["DATE"] == prod_from_date].copy()
                df_merged = df.merge(df_date, on=["REAL"], how="inner")
                for vec in sumvecs:
                    df_merged[vec] = df_merged[f"{vec}_x"] - df_merged[f"{vec}_y"]
                df = df_merged[["REAL", "DATE_x"] + sumvecs].rename(
                    {"DATE_x": "DATE"}, axis=1
                )

        if prod_until_date is not None:
            df = df[df["DATE"] <= prod_until_date]

        return df

    def get_dataframe_melted(
        self,
        well_sumvec: str,
        prod_from_date: Union[datetime.datetime, None],
        prod_until_date: Union[datetime.datetime, None],
    ) -> pd.DataFrame:
        """Returns a dataframe on long form consisting of these columns:
        * WELL
        * well_sumvec (f.ex WOPT)
        * ENSEMBLE
        """
        sumvecs = [f"{well_sumvec}:{well}" for well in self._wells]
        df = self._smry[["REAL", "DATE"] + sumvecs]
        max_date = df["DATE"].max()
        min_date = df["DATE"].min()

        if prod_until_date is None:
            prod_until_date = max_date
        else:
            # Set prod_until_date to min_date or max_date if it is outside the
            # ensemble date range
            prod_until_date = max(min(prod_until_date, max_date), min_date)

        df = df[df["DATE"] == prod_until_date]

        # If prod_from_date is None, do nothing
        if prod_from_date is not None:
            # Set prod_from_date to min_date or max_date if it is outside the
            # ensemble date range
            prod_from_date = max(min(prod_from_date, max_date), min_date)

            # Subtract the production at the prod_from_date
            df_date = self._smry[["REAL", "DATE"] + sumvecs]
            df_date = df_date[df_date["DATE"] >= prod_from_date]
            df_date = df_date[df_date["DATE"] == df_date["DATE"].min()]
            df_merged = df.merge(df_date, on=["REAL"], how="inner")
            for vec in sumvecs:
                df_merged[vec] = df_merged[f"{vec}_x"] - df_merged[f"{vec}_y"]
            df = df_merged[["REAL"] + sumvecs]

        df_melted = pd.melt(
            df, value_vars=sumvecs, var_name="WELL", value_name=well_sumvec
        )
        df_melted["WELL"] = df_melted.agg(
            lambda x: f"{x['WELL'].split(':')[1]}", axis=1
        )
        df_melted["ENSEMBLE"] = self._ensemble_name
        return df_melted

    def get_node_info(
        self,
        node: str,
        pressure_plot_mode: PressurePlotMode,
        real: int,
        node_type: NodeType = NodeType.WELL,
    ) -> Dict[str, Any]:
        """Returns a list of dictionaries containing the network nodes
        ending in the input node, with from_date and end_date. If end_date
        is None it means that the network is valid for the rest of the
        simulations.
        Dict[str, Any]
        The output has the form:
        {
            "name": "A1",
            "type: "well",
            "ctrlmode_sumvec": "WMCTL:A1",
            "networks: [
                {
                    "start_date": "2018-01-01",
                    "end_date": "2018-05-01",
                    "nodes": [
                        {
                            "name": "A1",
                            "label": "THP"
                            "pressure_sumvec": "WTHP:A1"
                            "type": "well"
                        },
                        ...
                    ],
                },
                ...
            ]
        }
        """

        gruptree_df = self._gruptree_model.dataframe

        if gruptree_df.empty or (
            pressure_plot_mode == PressurePlotMode.MEAN
            and not self._gruptree_model.gruptrees_are_equal_over_reals
        ):
            # If there is no gruptree OR
            # the plot mode is MEAN and the gruptrees are different over
            # realization, we return only data for the current node.
            nodes = [_get_node_field(node_type, node)]
            if node_type == NodeType.WELL:
                nodes.append(_get_node_field(NodeType.WELL_BH, node))
            return {
                "name": node,
                "type": node_type.value,
                "ctrlmode_sumvec": _get_ctrlmode_sumvec(node_type, node),
                "networks": [
                    {
                        "start_date": self._smry["DATE"].min(),
                        "end_date": None,
                        "nodes": nodes,
                    }
                ],
            }

        if (
            pressure_plot_mode == PressurePlotMode.SINGLE_REAL
            and not self._gruptree_model.gruptrees_are_equal_over_reals
        ):
            # If the plot mode is SINGLE_REAL and gruptrees are
            # different over realizations, then we need to filter
            # the gruptree dataframe on realization
            gruptree_df = gruptree_df[gruptree_df["REAL"] == real]

        node_networks: List[Dict[str, Any]] = []
        prev_nodelist: List[Dict[str, Any]] = []
        for date, df in gruptree_df.groupby("DATE"):
            nodelist = _get_nodelist(df, node_type, node)
            remaining_dates = gruptree_df[gruptree_df["DATE"] > date].DATE
            if remaining_dates.empty:
                next_date = None
            else:
                next_date = remaining_dates.min()

            if nodelist != prev_nodelist:
                node_networks.append(
                    {"start_date": date, "end_date": next_date, "nodes": nodelist}
                )
                prev_nodelist = nodelist
            else:
                if node_networks:
                    node_networks[-1]["end_date"] = next_date
        return {
            "name": node,
            "type": node_type.value,
            "ctrlmode_sumvec": _get_ctrlmode_sumvec(node_type, node),
            "networks": node_networks,
        }


def _get_nodelist(
    df: pd.DataFrame, node_type: NodeType, node: str
) -> List[Dict[str, str]]:
    """Returns a list of node dictionaries ending up in the input
    node. The input dataframe has only one date.

    The function is recursively getting all the nodes ending up in
    the node which the function was called from the outside.

    For wells, both THP and BHP nodes are added
    """
    if node == "FIELD":
        return [_get_node_field(node_type, node)]

    child_row = df[df.CHILD == node]

    if child_row.empty:
        return []
    if child_row.shape[0] > 1:
        raise ValueError(
            f"There can be maximum one row per child per date: {child_row}"
        )
    if is_terminal_node(node, child_row):
        return [_get_node_field(NodeType.GROUP, node)]
    parent = child_row.PARENT.values[0]
    nodelist = [_get_node_field(node_type, node)]
    if node_type == NodeType.WELL:
        nodelist.append(_get_node_field(NodeType.WELL_BH, node))
    return nodelist + _get_nodelist(df, NodeType.GROUP, parent)


def is_terminal_node(node: str, row: pd.Series) -> bool:
    """Checks if the input node is a terminal node in the network.
    FIELD is always a terminal node, but other nodes are terminal if they
    have a value in the TERMINAL_PRESSURE column.
    """
    if node == "FIELD":
        return True
    if (
        "TERMINAL_PRESSURE" in row
        and row["TERMINAL_PRESSURE"].values[0] is not None
        and not np.isnan(row["TERMINAL_PRESSURE"].values[0])
    ):
        return True
    return False


def _get_node_field(node_type: NodeType, node: str) -> Dict[str, str]:
    """Returns a dictionary with info about a single node:
    * Name
    * Label to be used in pressure plot
    * Type: well, well-bh, group or terminal-node
    """
    if node_type == NodeType.GROUP:
        return {
            "name": node,
            "label": node,
            "type": node_type,
            "pressure": f"GPR:{node}",
        }
    if node_type == NodeType.WELL:
        return {
            "name": node,
            "label": "THP",
            "type": node_type,
            "pressure": f"WTHP:{node}",
        }
    if node_type == NodeType.WELL_BH:
        return {
            "name": node,
            "label": "BHP",
            "type": node_type,
            "pressure": f"WBHP:{node}",
        }
    raise ValueError(f"Node type {node_type} not implemented.")


def _get_ctrlmode_sumvec(node_type: NodeType, node: str) -> str:
    """Returns the control mode sumvec for a given node type
    and node name. Only production network implemented so far.
    """
    if node == "FIELD":
        return "FMCTP"
    if node_type == NodeType.WELL:
        return f"WMCTL:{node}"
    if node_type == NodeType.GROUP:
        return f"GMCTP:{node}"
    raise ValueError(f"Node type {node_type} not implemented")



from pathlib import Path
from typing import Dict, List, Optional, Sequence, Tuple, TypedDict, Union
from uuid import uuid4

import numpy as np
import pandas as pd
import yaml
from webviz_subsurface_components import (
    ExpressionInfo,
    ExternalParseData,
    VariableVectorMapInfo,
    VectorCalculator,
    VectorDefinition,
)

from webviz_subsurface._providers import EnsembleSummaryProvider, Frequency

from .vector_selector import (
    add_vector_to_vector_selector_data,
    is_vector_name_in_vector_selector_data,
)

# JSON Schema for predefined expressions configuration
# Used as schema input for json_schema.validate()
PREDEFINED_EXPRESSIONS_JSON_SCHEMA = {
    "type": "object",
    "additionalProperties": {
        "type": "object",
        "required": ["expression", "variableVectorMap"],
        "properties": {
            "expression": {"type": "string", "minLength": 1},
            "description": {
                "type": "string",
                "maxLength": VectorCalculator.max_description_length,
            },
            "variableVectorMap": {
                "type": "object",
                "minProperties": 1,
                "patternProperties": {"^[a-zA-Z]$": {"type": "string"}},
                "additionalProperties": False,
            },
        },
        "additionalProperties": False,
    },
}


class ConfigExpressionDataBase(TypedDict):
    """Base definition for configuration of pre-defined calculated expressions

    `Description`:
    * Simplified data type to pre-define expressions for user.
    * The ConfigExpressionData instance name is the name of the expression - i.e.
    when using a dictionary of ConfigExpressionData the expression name is the dict key

    `Required keys`:
    * expression: str, mathematical expression
    * variableVectorMap: Dict[str,str], Dictionary with {key, value} = {variableName, vectorName}
    """

    expression: str
    variableVectorMap: Dict[str, str]


class ConfigExpressionData(ConfigExpressionDataBase, total=False):
    """
    Type definition for configuration of pre-defined calculated expressions

    `Description`:
    Dictionary type for pre-defined expressions for user.

    `Required keys`:
    * expression: str, mathematical expression
    * variableVectorMap: Dict[str,str], Dictionary with {key, value} = {variableName, vectorName}

    `Non-required keys`:
    * description: str, description of mathematical expression
    """

    description: str


def validate_predefined_expression(
    expression: ExpressionInfo, vector_data: list
) -> Tuple[bool, str]:
    """
    Validates predefined expressions for usage in vector calculator

    Predefined expressions can be defined in configuration file. Validation will ensure valid
    mathematical expression parsing and matching equation variables and variable vector map.
    It will also verify provided vector names in map is represented in provided vector data

    Inputs:
    * expression: Predefined expression
    * vector_data: Vector data

    Returns:
    * Tuple of valid state and validation message. Validation message is empty for valid expression

    """
    parsed_expression: ExternalParseData = VectorCalculator.external_parse_data(
        expression
    )
    expr: str = expression["expression"]
    name: str = expression["name"]

    # Validate expression string
    if not parsed_expression["isValid"]:
        parse_message = parsed_expression["message"]
        message = (
            f'Invalid mathematical expression {expr} in predefined expression "{name}".'
            f"{parse_message}."
        )
        return False, message

    # Match variables in expression string and variable names in map
    expression_variables = parsed_expression["variables"]
    map_variables = [elm["variableName"] for elm in expression["variableVectorMap"]]
    if set(expression_variables) != set(map_variables):
        message = (
            f"Variables {map_variables} in variableVectorMap is inconsistent with variables "
            f'{expression_variables} in equation "{expr}" for predefined expression "{name}"'
        )
        return False, message

    # Validate vector names
    variable_vector_dict = VectorCalculator.variable_vector_dict(
        expression["variableVectorMap"]
    )
    invalid_vectors: List[str] = []
    for vector_name in variable_vector_dict.values():
        if not is_vector_name_in_vector_selector_data(vector_name, vector_data):
            invalid_vectors.append(vector_name)
    if len(invalid_vectors) > 1:
        message = (
            f'Vector names {invalid_vectors} in predefined expression "{name}" are not'
            f" represented in vector data"
        )
        return False, message
    if len(invalid_vectors) > 0:
        message = (
            f'Vector name {invalid_vectors} in predefined expression "{name}" is not'
            f" represented in vector data"
        )
        return False, message

    return True, ""


def variable_vector_map_from_dict(
    variable_vector_dict: Dict[str, str]
) -> List[VariableVectorMapInfo]:
    variable_vector_map: List[VariableVectorMapInfo] = []
    for variable in variable_vector_dict:
        variable_vector_map.append(
            {
                "variableName": variable,
                "vectorName": [variable_vector_dict[variable]],
            }
        )
    return variable_vector_map


def expressions_from_config(
    predefined_expressions: Optional[Path],
) -> List[ExpressionInfo]:
    output: List[ExpressionInfo] = []

    if predefined_expressions is None:
        return output

    expressions: Dict[str, ConfigExpressionData] = yaml.safe_load(
        predefined_expressions.read_text()
    )

    for expression in expressions:
        expression_dict = {
            "name": expression,
            "expression": expressions[expression]["expression"],
            "id": str(uuid4()),
            "variableVectorMap": variable_vector_map_from_dict(
                expressions[expression]["variableVectorMap"]
            ),
            "isValid": False,  # Set False and validate in seperate operation
            "isDeletable": False,
        }
        if "description" in expressions[expression]:
            expression_dict["description"] = expressions[expression]["description"]

        output.append(expression_dict)
    return output


def get_expression_from_name(
    name: str, expressions: List[ExpressionInfo]
) -> Union[ExpressionInfo, None]:
    for expr in expressions:
        if name == expr["name"]:
            return expr
    return None


def get_vector_definitions_from_expressions(
    expressions: List[ExpressionInfo],
) -> Dict[str, VectorDefinition]:
    """
    Get vector definitions for vector selector from list of calculated expressions.

    VectorSelector has VectorDefinitions which is utilized for calculated expressions.

    `VectorDefinitions:`
        key: str, vector name
        value: {
            type: str, defined vector type
            description: str, description of vector
        }

    `Note:`
    Uses expression str as description if optional expression description str does not exist.
    """

    output: Dict[str, VectorDefinition] = {}
    for expression in expressions:
        name = expression["name"]
        key = name.split(":")[0]
        vector_type = "calculated"
        description = (
            expression["expression"]
            if not "description" in expression
            else expression["description"]
        )
        output[key] = VectorDefinition(type=vector_type, description=description)
    return output


def get_selected_expressions(
    expressions: List[ExpressionInfo], selected_names: List[str]
) -> List[ExpressionInfo]:
    selected: List[ExpressionInfo] = []
    for name in selected_names:
        selected_expression: ExpressionInfo = next(
            (elm for elm in expressions if elm["name"] == name), None
        )
        if selected_expression is not None and selected_expression not in selected:
            selected.append(selected_expression)
    return selected


def get_calculated_vector_df(
    expression: ExpressionInfo, smry: pd.DataFrame, ensembles: List[str]
) -> pd.DataFrame:
    columns = ["REAL", "ENSEMBLE", "DATE"]

    name: str = expression["name"]
    expr: str = expression["expression"]

    var_vec_dict: Dict[str, str] = VectorCalculator.variable_vector_dict(
        expression["variableVectorMap"]
    )
    vector_names = var_vec_dict.values()

    # Retreive vectors for calculating expression - filtered on ensembles
    df = smry[columns + list(vector_names)].copy()
    df = df[df["ENSEMBLE"].isin(ensembles)]

    values: Dict[str, np.ndarray] = {}
    for variable, vector in var_vec_dict.items():
        values[variable] = df[vector].values

    evaluated_expr = VectorCalculator.evaluate_expression(expr, values)
    if evaluated_expr is not None:
        df[name] = evaluated_expr

    return df[columns + [name]]


def create_calculated_vector_df(
    expression: ExpressionInfo,
    provider: EnsembleSummaryProvider,
    realizations: Optional[Sequence[int]],
    resampling_frequency: Optional[Frequency],
) -> pd.DataFrame:
    """Create dataframe with calculated vector from expression

    If expression is not successfully evaluated, empty dataframe is returned

    `Return:`
    * Dataframe with calculated vector data made form expression - columns:\n
        ["DATE","REAL", calculated_vector]
    * Return empty dataframe if expression evaluation returns None
    """
    name: str = expression["name"]
    expr: str = expression["expression"]

    variable_vector_dict: Dict[str, str] = VectorCalculator.variable_vector_dict(
        expression["variableVectorMap"]
    )
    vector_names = list(variable_vector_dict.values())

    # Retrieve data for vectors in expression
    vectors_df = provider.get_vectors_df(
        vector_names, resampling_frequency, realizations
    )

    values: Dict[str, np.ndarray] = {}
    for variable, vector in variable_vector_dict.items():
        values[variable] = vectors_df[vector].values

    evaluated_expression = VectorCalculator.evaluate_expression(expr, values)
    if evaluated_expression is not None:
        vectors_df[name] = evaluated_expression
        return vectors_df[["DATE", "REAL", name]]

    return pd.DataFrame()


def get_calculated_units(
    expressions: List[ExpressionInfo],
    units: pd.Series,
) -> pd.Series:
    # TODO Update expression handling
    # Future: check equal equal unit on each vector:
    # - if equal unit on + or -: x[m]+y[m] = [m]
    # - If unequal unit on + or -: Set unit "mixed"?
    # - *, / or ^: perform operators on units
    #
    # Utilize ./_datainput/units.py and/or ./_datainput/eclipse_unit.py
    #
    #
    # Now: parse expression str with VectorCalculator.parser.parse()
    # if valid, do string replace with units from smry_meta
    calculated_units: pd.Series = pd.Series()

    for expression in expressions:
        try:
            # Parse only for validation
            VectorCalculator.parser.parse(expression["expression"])
            unit_expr: str = expression["expression"]
            for elm in expression["variableVectorMap"]:
                unit_expr = unit_expr.replace(
                    elm["variableName"], units[elm["vectorName"][0]]
                )

            calculated_units[expression["name"]] = unit_expr
        except ValueError:
            continue
    return calculated_units


def add_calculated_vector_to_vector_selector_data(
    vector_selector_data: list,
    vector_name: str,
    description: Optional[str] = None,
) -> None:
    """Add calculated vector name and descritpion to vector selector data

    Description is optional, and will be added at last node
    """
    description_str = description if description is not None else ""
    add_vector_to_vector_selector_data(
        vector_selector_data=vector_selector_data,
        vector=vector_name,
        description=description_str,
        description_at_last_node=True,
    )


def add_expressions_to_vector_selector_data(
    vector_selector_data: list, expressions: List[ExpressionInfo]
) -> None:
    """Add expressions to vector selector data

    Adds calculated vector name into node structure. Adds expression
    description if existing.
    """
    for expression in expressions:
        if not expression["isValid"]:
            continue

        name = expression["name"]
        description = None
        if "description" in expression.keys():
            description = expression["description"]

        add_calculated_vector_to_vector_selector_data(
            vector_selector_data, name, description
        )



# pylint: disable=no-name-in-module
from webviz_config.plugins import ReservoirSimulationTimeSeriesOneByOne


def test_reservoir_simulation_timeseries_onebyone(
    dash_duo, app, shared_settings
) -> None:
    plugin = ReservoirSimulationTimeSeriesOneByOne(
        app,
        shared_settings["SENS_SETTINGS"],
        ensembles=shared_settings["SENS_ENSEMBLES"],
        initial_vector="FOPT",
    )
    app.layout = plugin.layout
    dash_duo.start_server(app)
    assert dash_duo.get_logs() == []



import datetime

# DOCS: https://docs.python.org/3/library/datetime.html#strftime-strptime-behavior


def from_str(date_str: str) -> datetime.datetime:
    return datetime.datetime.strptime(date_str, "%Y-%m-%d")


def to_str(date: datetime.datetime) -> str:
    if date.hour != 0 or date.minute != 0 or date.second != 0 or date.microsecond != 0:
        raise ValueError(
            f"Invalid date resolution, expected no data for hour, minute, second"
            f" or microsecond for {str(date)}"
        )
    return date.strftime("%Y-%m-%d")



from typing import Callable, Dict, List, Union

import webviz_core_components as wcc
from dash import html


def data_selectors_view(
    get_uuid: Callable,
    ensemble_names: List[str],
    data_column_names: List[str],
    parameter_names: List[str],
    initial_data: Dict,
) -> html.Div:
    return wcc.Selectors(
        label="Selectors",
        children=[
            dropdown_for_plotly_data(
                uuid=get_uuid("data_selectors"),
                source="table",
                data_attribute="x",
                title="X-value",
                options=[{"label": col, "value": col} for col in data_column_names],
                value=initial_data.get("x", data_column_names[0]),
            ),
            dropdown_for_plotly_data(
                uuid=get_uuid("data_selectors"),
                source="table",
                data_attribute="y",
                title="Y-value",
                options=[{"label": col, "value": col} for col in data_column_names],
                value=initial_data.get("y", data_column_names[-1]),
            ),
            dropdown_for_plotly_data(
                uuid=get_uuid("data_selectors"),
                source="table",
                data_attribute="ensemble",
                title="Ensemble",
                options=[{"label": col, "value": col} for col in ensemble_names],
                value=initial_data.get("ensembles", [ensemble_names[0]]),
                multi=True,
            ),
            dropdown_for_plotly_data(
                uuid=get_uuid("data_selectors"),
                source="parameter",
                data_attribute="parameter",
                title="Color by",
                options=[{"label": col, "value": col} for col in parameter_names],
                clearable=True,
                placeholder="Default ensemble color",
                value=initial_data.get("parameter"),
            ),
        ],
    )


# pylint: disable=too-many-arguments
def dropdown_for_plotly_data(
    uuid: str,
    source: str,
    data_attribute: str,
    title: str,
    options: List[Dict],
    value: Union[List, str] = None,
    flex: int = 1,
    placeholder: str = "Select...",
    multi: bool = False,
    clearable: bool = False,
) -> html.Div:
    return html.Div(
        style={"flex": flex},
        children=[
            wcc.Dropdown(
                label=title,
                id={"id": uuid, "data_attribute": data_attribute, "source": source},
                options=options,
                value=value,
                clearable=clearable,
                placeholder=placeholder,
                multi=multi,
            ),
        ],
    )


def color_scale_dropdown(
    uuid: str,
    title: str,
    options: List[Dict],
    value: Union[List, str] = None,
    flex: int = 1,
    placeholder: str = "Select...",
    multi: bool = False,
    clearable: bool = False,
) -> html.Div:
    return html.Div(
        style={"flex": flex},
        children=[
            wcc.Dropdown(
                label=title,
                id=uuid,
                options=options,
                value=value,
                clearable=clearable,
                placeholder=placeholder,
                multi=multi,
            ),
        ],
    )



# pylint: disable=no-name-in-module
from webviz_config.plugins import RftPlotter
from webviz_config.testing import WebvizComposite


def test_rft_plotter(
    _webviz_duo: WebvizComposite, shared_settings, testdata_folder
) -> None:
    plugin = RftPlotter(
        shared_settings["HM_SETTINGS"],
        ensembles=shared_settings["HM_ENSEMBLES"],
        formations=testdata_folder
        / "01_drogon_ahm"
        / "realization-0"
        / "iter-0"
        / "share"
        / "results"
        / "tables"
        / "formations_res_only.csv",
        faultlines=testdata_folder
        / "01_drogon_ahm"
        / "realization-0"
        / "iter-0"
        / "share"
        / "results"
        / "polygons"
        / "toptherys--gl_faultlines_extract_postprocess.csv",
    )

    _webviz_duo.start_server(plugin)

    _webviz_duo.toggle_webviz_settings_drawer()
    _webviz_duo.toggle_webviz_settings_group(
        plugin.view("map-view").settings_group_unique_id("map-settings")
    )
    # Using str literals directly, not IDs from the plugin as intended because
    # the run test did not accept the imports

    my_component_id = _webviz_duo.view_settings_group_unique_component_id(
        "map-view", "map-settings", "map-ensemble"
    )
    _webviz_duo.wait_for_contains_text(my_component_id, "iter-0")
    assert _webviz_duo.get_logs() == []
